# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""HTML and the base HtmlView."""

import abc
import functools
import html as html_lib
import inspect
import typing
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Union

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.views import base

NestableStr = Union[
    str,
    Sequence[Union[str, None, Sequence[Optional[str]]]],
    None,
]

NodeFilter = base.NodeFilter
NodeColor = Callable[
    [
        object_utils.KeyPath,    # The path to the value.
        Any,        # Current value.
        Any,        # Parent value
    ],
    Optional[str]   # The color of the node.
]


class Html(base.Content):
  """HTML with consolidated CSS and Scripts.

  Example::

  .. code-block:: python

    def foo() -> pg.Html:
      s = pg.Html()
      s.add_style('div.foo { color: red; }')
      s.add_script('function myFoo() { console.log("foo");}')
      s.write('<div class="foo">Foo</div>')
      return s

    def bar() -> pg.Html:
      s = pg.Html()
      s.add_style('div.bar { color: green; }')
      s.add_script('function myBar() { console.log("bar");}')
      s.write('<div class="bar">')
      s.write(foo())
      s.write('</div>')
      return s

    html = bar.html_str()

    This will output::

     <html>
     <head>
     <style>
     div.bar { color: green; }
     div.foo { color: red; }
     </style>
     <script>
     function myBar() { console.log("bar");}
     function myFoo() { console.log("foo");}
     </script>
     </head>
     <body><div class="bar"><div class="foo">Foo</div></div></body></html>

  """

  WritableTypes = Union[    # pylint: disable=invalid-name
      str,
      'Html',
      Callable[[], Union[str, 'Html', None]],
      None
  ]

  class Scripts(base.Content.SharedParts):
    """Shared script definitions in the HEAD section."""

    @functools.cached_property
    def content(self) -> str:
      if self.parts:
        code = '\n'.join([inspect.cleandoc(v) for v in self.parts.keys()])
        return f'<script>\n{code}\n</script>'
      return ''

  class ScriptFiles(base.Content.SharedParts):
    """Shared script files to link to in the HEAD section."""

    @functools.cached_property
    def content(self) -> str:
      return '\n'.join(
          [f'<script src="{url}"></script>' for url in self.parts.keys()]
      )

  class Styles(base.Content.SharedParts):
    """Shared style definitions in the HEAD section."""

    @functools.cached_property
    def content(self) -> str:
      if self.parts:
        styles = '\n'.join([inspect.cleandoc(v) for v in self.parts.keys()])
        return f'<style>\n{styles}\n</style>'
      return ''

  class StyleFiles(base.Content.SharedParts):
    """Shared style files to link to in the HEAD section."""

    @functools.cached_property
    def content(self) -> str:
      return '\n'.join(
          [
              f'<link rel="stylesheet" href="{url}">'
              for url in self.parts.keys()
          ]
      )

  def __init__(  # pylint: disable=useless-super-delegation
      self,
      *content: WritableTypes,
      style_files: Optional[Iterable[str]] = None,
      styles: Optional[Iterable[str]] = None,
      script_files: Optional[Iterable[str]] = None,
      scripts: Optional[Iterable[str]] = None,
  ) -> None:
    """Constructor.

    Args:
      *content: One or multiple body part (str, Html, lambda, None) of the HTML.
      style_files: URLs for external styles to include.
      styles: CSS styles to include.
      script_files: URLs for external scripts to include.
      scripts: JavaScript scripts to include.
    """
    super().__init__(
        *content,
        style_files=Html.StyleFiles(*(style_files or [])),
        styles=Html.Styles(*(styles or [])),
        script_files=Html.ScriptFiles(*(script_files or [])),
        scripts=Html.Scripts(*(scripts or [])),
    )

  def _repr_html_(self) -> str:
    return self.to_str()

  @property
  def styles(self) -> 'Html.Styles':
    """Returns the styles to include in the HTML."""
    return self._shared_parts['styles']

  @property
  def style_files(self) -> 'Html.StyleFiles':
    """Returns the style files to link to."""
    return self._shared_parts['style_files']

  @property
  def scripts(self) -> 'Html.Scripts':
    """Returns the scripts to include in the HTML."""
    return self._shared_parts['scripts']

  @property
  def script_files(self) -> 'Html.ScriptFiles':
    """Returns the script files to link to."""
    return self._shared_parts['script_files']

  @property
  def head_section(self) -> str:
    """Returns the head section."""
    return '\n'.join(
        [
            v for v in [
                '<head>', self.style_section, self.script_section, '</head>']
            if v
        ]
    )

  @property
  def style_section(self) -> str:
    """Returns the style section."""
    return '\n'.join(
        [
            v for v in [self.style_files.content, self.styles.content]
            if v
        ]
    )

  @property
  def script_section(self) -> str:
    """Returns the script section."""
    return '\n'.join(
        [
            v for v in [self.script_files.content, self.scripts.content]
            if v
        ]
    )

  @property
  def body_section(self) -> str:
    """Returns the body section."""
    return f'<body>\n{self.content}\n</body>'

  #
  # Methods for adding shared parts and writing HTML content.
  #

  def add_style(self, *css: str) -> 'Html':
    """Adds CSS styles to the HTML."""
    self.styles.add(*css)
    return self

  def add_script(self, *js: str) -> 'Html':
    """Adds JavaScript scripts to the HTML."""
    self.scripts.add(*js)
    return self

  def add_style_file(self, *url: str) -> 'Html':
    """Adds a style file to the HTML."""
    self.style_files.add(*url)
    return self

  def add_script_file(self, *url: str) -> 'Html':
    """Adds a script file to the HTML."""
    self.script_files.add(*url)
    return self

  def to_str(
      self,
      *,
      content_only: bool = False,
      **kwargs
  ) -> str:
    """Returns the HTML str.

    Args:
      content_only: If True, only the content will be returned.
      **kwargs: Additional keyword arguments passed from the user that 
        will be ignored.

    Returns:
      The generated HTML str.
    """
    if content_only:
      return self.content
    return '\n'.join(
        [
            v for v in ['<html>', self.head_section,
                        self.body_section, '</html>'] if v
        ]
    )

  @classmethod
  def from_value(
      cls,
      value: WritableTypes,
      copy: bool = False
  ) -> Union['Html', None]:
    return typing.cast(
        Html, super().from_value(value, copy=copy)
    )

  #
  # Helper methods for creating templated Html objects.
  #

  @classmethod
  def element(
      cls,
      tag: str,
      inner_html: Optional[List[WritableTypes]] = None,
      *,
      options: Union[str, Iterable[str], None] = None,
      css_class: NestableStr = None,
      style: Union[str, Dict[str, Any], None] = None,
      **properties
  ) -> 'Html':
    """Creates an HTML element.

    Args:
      tag: The HTML tag name.
      inner_html: The inner HTML of the element.
      options: Positional options that will be added to the element. E.g. 'open'
        for `<details open>`.
      css_class: The CSS class name or a list of CSS class names.
      style: A single CSS style string or a dictionary of CSS properties.
      **properties: Keyword arguments for HTML properties. For properties with
        underscore in the name, the underscore will be replaced by dash in the
        generated HTML. E.g. `background_color` will be converted to
        `background-color`.

    Returns:
      The opening tag of an HTML element.
    """
    s = cls()

    # Write the open tag.
    css_class = cls.concate(css_class)
    options = cls.concate(options)
    style = cls.style_str(style)
    s.write(
        f'<{tag}',
        f' {options}' if options else None,
        f' class="{css_class}"' if css_class else None,
        f' style="{style}"' if style else None,
    )
    for k, v in properties.items():
      if v is not None:
        s.write(f' {k.replace("_", "-")}="{v}"')
    s.write('>')

    # Write the inner HTML.
    if inner_html:
      for child in inner_html:
        s.write(child)

    # Write the closing tag.
    s.write(f'</{tag}>')
    return s

  @classmethod
  def escape(cls, s: WritableTypes) -> WritableTypes:
    """Escapes an HTML writable object."""
    if s is None:
      return None

    if callable(s):
      s = s()

    if isinstance(s, str):
      return html_lib.escape(s)
    else:
      assert isinstance(s, Html), s
      return Html(html_lib.escape(s.content)).write(
          s, shared_parts_only=True
      )

  @classmethod
  def concate(
      cls, nestable_str: NestableStr, separator: str = ' '
  ) -> Optional[str]:
    """Concates the string nodes in a nestable object."""
    if isinstance(nestable_str, str):
      return nestable_str
    elif isinstance(nestable_str, list):
      flattened = [cls.concate(s) for s in nestable_str]
      flattened = [s for s in flattened if s is not None]
      return separator.join(flattened) if flattened else None
    else:
      assert nestable_str is None, nestable_str
      return None

  @classmethod
  def style_str(
      cls, style: Union[str, Dict[str, Any], None],
  ) -> Optional[str]:
    """Gets a string representing an inline CSS style.

    Args:
      style: A single CSS style string, or a dictionary for CSS properties.
        When dictionary form is used, underscore in the key name will be
        replaced by dash in the generated CSS style string.
        For example, `background_color` will be converted to `background-color`.

    Returns:
      A CSS style string or None if no CSS property is provided.
    """
    if not style:
      return None
    if isinstance(style, str):
      return style
    else:
      assert isinstance(style, dict), style
      return ''.join(
          [
              f'{k.replace("_", "-")}:{v};'
              for k, v in style.items() if v is not None
          ]
      ) or None


# Allow automatic conversion from str to Html.
pg_typing.register_converter(str, Html, convert_fn=Html.from_value)


class HtmlView(base.View):
  """Base class for HTML views."""

  class Extension(base.View.Extension):
    """Base class for HtmlView extensions."""

    def _html_style(self) -> List[str]:
      """Returns additional CSS style to add for this extension.

      Subclasses can override this method to add additional CSS style to the
      rendered HTML.
      """
      return []

    def _html_element_class(self) -> List[str]:
      """Returns the CSS classes for the rendered element for this node.

      Subclasses can override this method to add CSS classes to the
      rendered element of this object.
      """
      return [
          object_utils.camel_to_snake(self.__class__.__name__, '-')
      ]

    def to_html(
        self,
        *,
        name: Optional[str] = None,
        root_path: Optional[object_utils.KeyPath] = None,
        view_id: str = 'html-tree-view',
        **kwargs
    ) -> Html:
      """Returns the HTML representation of the object.

      Args:
        name: The name of the object.
        root_path: The root path of the object.
        view_id: The ID of the view to render the value.
          See `pg.views.HtmlView.dir()` for all available HTML view IDs.
        **kwargs: View-specific keyword arguments passed to `pg.to_html`, wich
          will be used to construct/override `HtmlView` settings.

      Returns:
        An rendered HTML.
      """
      return to_html(
          self, name=name, root_path=root_path, view_id=view_id, **kwargs
      )

    def to_html_str(
        self,
        *,
        name: Optional[str] = None,
        root_path: Optional[object_utils.KeyPath] = None,
        view_id: str = 'html-tree-view',
        content_only: bool = False,
        **kwargs
    ) -> str:
      """Returns the HTML str of the object.

      Args:
        name: The name of the object.
        root_path: The root path of the object.
        view_id: The ID of the view to render the value.
          See `pg.views.HtmlView.dir()` for all available HTML view IDs.
        content_only: If True, only the content will be returned.
        **kwargs: View-specific keyword arguments passed to `pg.to_html`, wich
          will be used to construct/override `HtmlView` settings.

      Returns:
        An rendered HTML str.
      """
      return to_html_str(
          self, name=name, root_path=root_path,
          view_id=view_id, content_only=content_only, **kwargs
      )

    def _repr_html_(self) -> str:
      return self.to_html_str()

  @abc.abstractmethod
  def render(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      root_path: Optional[object_utils.KeyPath] = None,
      **kwargs
  ) -> Html:
    """Renders the input value into an HTML object."""


def to_html(
    value: Any,
    *,
    name: Optional[str] = None,
    root_path: Optional[object_utils.KeyPath] = None,
    view_id: str = 'html-tree-view',
    **kwargs
) -> Html:
  """Returns the HTML representation of a value.

  Args:
    value: The value to render.
    name: The name of the value.
    root_path: The root path of the value.
    view_id: The ID of the view to render the value.
      See `pg.views.HtmlView.dir()` for all available HTML view IDs.
    **kwargs: Additional keyword arguments passed from `pg.to_html`, wich
        will be passed to the `HtmlView.render_xxx()` (thus
        `Extension._html_xxx()`) methods.

  Returns:
    The rendered HTML.
  """
  content = base.view(
      value,
      name=name,
      root_path=root_path,
      view_id=view_id,
      **kwargs
  )
  assert isinstance(content, Html), content
  return content


def to_html_str(
    value: Any,
    *,
    name: Optional[str] = None,
    root_path: Optional[object_utils.KeyPath] = None,
    view_id: str = 'html-tree-view',
    content_only: bool = False,
    **kwargs
) -> str:
  """Returns a HTML str for a value.

  Args:
    value: The value to render.
    name: The name of the value.
    root_path: The root path of the value.
    view_id: The ID of the view to render the value.
      See `pg.views.HtmlView.dir()` for all available HTML view IDs.
    content_only: If True, only the content will be returned.
    **kwargs: Additional keyword arguments passed from `pg.to_html`, wich
        will be passed to the `HtmlView.render_xxx()` (thus
        `Extension._html_xxx()`) methods.

  Returns:
    The rendered HTML str.
  """
  return to_html(
      value,
      name=name,
      root_path=root_path,
      view_id=view_id,
      **kwargs
  ).to_str(content_only=content_only)




# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pg.Ref."""

import copy
import inspect
import pickle
from typing import Any
import unittest

from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import ref
from pyglove.core.symbolic.base import contains
from pyglove.core.symbolic.dict import Dict
from pyglove.core.symbolic.object import Object


class A(Object):
  x: int


class RefTest(unittest.TestCase):

  def test_basics(self):

    a = A(1)
    r = ref.Ref(a)
    self.assertIs(r.value, a)
    self.assertIs(r.infer(), a)
    self.assertEqual(ref.Ref(a), r)
    b = copy.copy(a)
    self.assertNotEqual(ref.Ref(b), r)

  def test_new(self):
    self.assertIsInstance(ref.Ref(1), int)
    self.assertIsInstance(ref.Ref(1.0), float)
    self.assertIsInstance(ref.Ref(True), bool)
    self.assertIsInstance(ref.Ref('abc'), str)

    r = ref.Ref(A(1))
    self.assertIsInstance(r, ref.Ref)
    self.assertIsInstance(r.infer(), A)
    self.assertEqual(ref.Ref(r), r)

  def test_type_check(self):

    class B:
      pass

    r = ref.Ref(A(1))
    self.assertIs(pg_typing.Object(A).apply(r), r)
    with self.assertRaisesRegex(
        TypeError, 'Expect .* but encountered .*'
    ):
      pg_typing.Object(A).apply(ref.Ref(B()))

  def test_partial_check(self):
    r = ref.Ref(A.partial())
    self.assertIs(pg_typing.Object(A).apply(r, allow_partial=True), r)
    with self.assertRaisesRegex(
        ValueError, '.* is not fully bound'
    ):
      pg_typing.Object(A).apply(r)

  def test_clone_basics(self):
    r = ref.Ref(A(1))
    self.assertIs(r.clone().infer(), r.infer())
    self.assertIs(r.clone(deep=True).infer(), r.infer())

    self.assertIs(copy.copy(r).infer(), r.infer())
    self.assertIs(copy.deepcopy(r).infer(), r.infer())

  def test_self_reference(self):

    class B(Object):
      x: Any
      y: Any

    b = B.partial(x=1)
    with self.assertRaisesRegex(
        NotImplementedError, 'Self-referential object is not supported.'
    ):
      b.rebind(y=ref.Ref(b))

  def test_repr(self):
    self.assertEqual(repr(ref.Ref(A(1))), 'Ref(A(x=1))')

  def test_str(self):
    self.assertEqual(
        str(ref.Ref(A(1))),
        inspect.cleandoc("""
        Ref(
          value = A(
            x = 1
          )
        )
        """))

  def test_to_json(self):
    with self.assertRaisesRegex(
        TypeError, '.* cannot be serialized at the moment'):
      ref.Ref(A(1)).to_json()

  def test_pickle(self):
    with self.assertRaisesRegex(
        TypeError, '.* cannot be pickled at the moment'):
      pickle.dumps(ref.Ref(A(1)))

  def test_maybe_ref(self):
    a = ref.maybe_ref(A(1))
    self.assertIsInstance(a, A)
    self.assertNotIsInstance(a, ref.Ref)

    b = Dict(x=a)
    x = ref.maybe_ref(b.x)
    self.assertIsInstance(x, ref.Ref)
    self.assertIs(x.value, a)

  def test_deref(self):
    class Foo(Object):
      x: int
      y: Any

    # Deref on non-ref value.
    a = Foo(1, 2)
    self.assertIs(ref.deref(a), a)
    self.assertEqual(ref.deref(a), Foo(1, 2))

    # Deref top-level Ref.
    a = ref.Ref(Foo(1, 2))
    self.assertIs(ref.deref(a), a.value)

    a = ref.Ref(Foo(1, ref.Ref(Foo(2, ref.Ref(Foo(3, 4))))))
    self.assertTrue(contains(a, type=ref.Ref))
    a_prime = ref.deref(a)
    self.assertIs(a_prime, a.value)
    self.assertIsInstance(a_prime.sym_getattr('y'), ref.Ref)
    self.assertIsInstance(a_prime.y.sym_getattr('y'), ref.Ref)

    # Deref entire tree.
    a = ref.Ref(Foo(1, ref.Ref(Foo(2, ref.Ref(Foo(3, 4))))))
    self.assertTrue(contains(a, type=ref.Ref))
    a_prime = ref.deref(a, recursive=True)
    self.assertIs(a_prime, a.value)
    self.assertFalse(contains(a_prime, type=ref.Ref))

  def test_to_html(self):

    def assert_style(html, expected):
      expected = inspect.cleandoc(expected).strip()
      actual = html.style_section.strip()
      if actual != expected:
        print(actual)
      self.assertEqual(actual.strip(), expected)

    def assert_content(html, expected):
      expected = inspect.cleandoc(expected).strip()
      actual = html.content.strip()
      if actual != expected:
        print(actual)
      self.assertEqual(actual.strip(), expected)

    class Foo(Object):
      x: Any

    assert_style(
        Foo(ref.Ref(Foo(1))).to_html(),
        """
        <style>
        /* Tooltip styles. */
        span.tooltip {
          visibility: hidden;
          white-space: pre-wrap;
          font-weight: normal;
          background-color: #484848;
          color: #fff;
          padding: 10px;
          border-radius: 6px;
          position: absolute;
          z-index: 1;
        }
        /* Summary styles. */
        details.pyglove summary {
          font-weight: bold;
          margin: -0.5em -0.5em 0;
          padding: 0.5em;
        }
        .summary_name {
          display: inline;
          padding: 0 5px;
        }
        .summary_title {
          display: inline;
        }
        .summary_name + div.summary_title {
          display: inline;
          color: #aaa;
        }
        .summary_title:hover + span.tooltip {
          visibility: visible;
        }
        /* Type-specific styles. */
        .pyglove.str .summary_title {
          color: darkred;
          font-style: italic;
        }
        /* Object key styles. */
        .object_key {
          margin-right: 0.25em;
        }
        .object_key:hover + .tooltip {
          visibility: visible;
          background-color: darkblue;
        }
        .object_key.str {
          color: gray;
          border: 1px solid lightgray;
          background-color: ButtonFace;
          border-radius: 0.2em;
          padding: 0.3em;
        }
        .object_key.int::before{
          content: '[';
        }
        .object_key.int::after{
          content: ']';
        }
        .object_key.int{
          border: 0;
          color: lightgray;
          background-color: transparent;
          border-radius: 0;
          padding: 0;
        }
        /* Simple value styles. */
        .simple_value {
          color: blue;
          display: inline-block;
          white-space: pre-wrap;
          padding: 0.2em;
          margin-top: 0.15em;
        }
        .simple_value.str {
          color: darkred;
          font-style: italic;
        }
        .simple_value.int, .simple_value.float {
          color: darkblue;
        }
        /* Complex value styles. */
        span.empty_container::before {
            content: '(empty)';
            font-style: italic;
            margin-left: 0.5em;
            color: #aaa;
        }
        /* Value details styles. */
        details.pyglove {
          border: 1px solid #aaa;
          border-radius: 4px;
          padding: 0.5em 0.5em 0;
          margin: 0.1em 0;
        }
        details.pyglove.special_value {
          margin-bottom: 0.75em;
        }
        details.pyglove[open] {
          padding: 0.5em 0.5em 0.5em;
        }
        .highlight {
          background-color: Mark;
        }
        .lowlight {
          opacity: 0.2;
        }
        </style>
        """
    )
    assert_content(
        Foo(ref.Ref(Foo(1))).to_html(
            use_inferred=False,
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove foo"><summary><div class="summary_title">Foo(...)</div></summary><div class="complex_value foo"><table><tr><td><span class="object_key str">x</span></td><td><div><details class="pyglove ref"><summary><div class="summary_title">Foo(...)</div></summary><div class="complex_value foo"><table><tr><td><span class="object_key str">x</span></td><td><div><span class="simple_value int">1</span></div></td></tr></table></div></details></div></td></tr></table></div></details>
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.symbolic.base."""

import copy
import inspect
from typing import Any
import unittest

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import base
from pyglove.core.symbolic.dict import Dict
from pyglove.core.symbolic.inferred import ValueFromParentChain
from pyglove.core.symbolic.object import Object


class FieldUpdateTest(unittest.TestCase):
  """Tests for `pg.symbolic.FieldUpdate`."""

  def test_basics(self):
    x = Dict(x=1)
    f = pg_typing.Field('x', pg_typing.Int())
    update = base.FieldUpdate(object_utils.KeyPath('x'), x, f, 1, 2)
    self.assertEqual(update.path, 'x')
    self.assertIs(update.target, x)
    self.assertIs(update.field, f)
    self.assertEqual(update.old_value, 1)
    self.assertEqual(update.new_value, 2)

  def test_format(self):
    self.assertEqual(
        base.FieldUpdate(
            object_utils.KeyPath('x'), Dict(x=1), None, 1, 2
        ).format(compact=True),
        'FieldUpdate(parent_path=, path=x, old_value=1, new_value=2)',
    )

    self.assertEqual(
        base.FieldUpdate(
            object_utils.KeyPath('a'), Dict(x=Dict(a=1)).x, None, 1, 2
        ).format(compact=True),
        'FieldUpdate(parent_path=x, path=a, old_value=1, new_value=2)',
    )

  def test_eq_ne(self):
    x = Dict()
    f = pg_typing.Field('x', pg_typing.Int())
    self.assertEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
    )

    # Targets are not the same instance.
    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), Dict(), f, 1, 2),
    )

    # Fields are not the same instance.
    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('b'), x, copy.copy(f), 1, 2),
    )

    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 0, 2),
    )

    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2),
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 1),
    )

    self.assertNotEqual(
        base.FieldUpdate(object_utils.KeyPath('a'), x, f, 1, 2), Dict()
    )


class HtmlTreeViewExtensionTest(unittest.TestCase):

  def assert_content(self, html, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = html.content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual.strip(), expected)

  def test_to_html(self):

    class Foo(Object):
      x: int
      y: Any = 'foo'
      z: pg_typing.Int().freeze(1)

    # Disable tooltip.
    self.assert_content(
        Foo(x=1, y='foo').to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False
        ),
        """
        <details open class="pyglove foo"><summary><div class="summary_title">Foo(...)</div></summary><div class="complex_value foo"><table><tr><td><span class="object_key str">x</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">y</span></td><td><div><span class="simple_value str">&#x27;foo&#x27;</span></div></td></tr></table></div></details>
        """
    )
    # Hide frozen and default values.
    self.assert_content(
        Foo(x=1, y='foo').to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            collapse_level=0,
            hide_frozen=True,
            hide_default_values=True
        ),
        """
        <details class="pyglove foo"><summary><div class="summary_title">Foo(...)</div></summary><div class="complex_value foo"><table><tr><td><span class="object_key str">x</span></td><td><div><span class="simple_value int">1</span></div></td></tr></table></div></details>
        """
    )
    # Use inferred values.
    x = Dict(x=Dict(y=ValueFromParentChain()), y=2)
    self.assert_content(
        x.x.to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            use_inferred=False
        ),
        """
        <details open class="pyglove dict"><summary><div class="summary_title">Dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">y</span></td><td><div><details class="pyglove value-from-parent-chain"><summary><div class="summary_title">ValueFromParentChain(...)</div></summary><div class="complex_value value-from-parent-chain"><span class="empty_container"></span></div></details></div></td></tr></table></div></details>
        """
    )
    self.assert_content(
        x.x.to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            use_inferred=True
        ),
        """
        <details open class="pyglove dict"><summary><div class="summary_title">Dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">y</span></td><td><div><span class="simple_value int">2</span></div></td></tr></table></div></details>
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common traits for Python objects.

This file defines interfaces for describing the common traits of a Python
object, for example, partiality (MaybePartial), functor (Functor).
"""

import abc
from typing import Any, Dict, Optional


class MaybePartial(metaclass=abc.ABCMeta):
  """Interface for classes whose instances can be partially constructed.

  A ``MaybePartial`` object is an object whose ``__init__`` method can accept
  ``pg.MISSING_VALUE`` as its argument values. All symbolic types (see
  :class:`pyglove.Symbolic`) implements this interface, as their symbolic
  attributes can be partially filled.

  Example::

    d = pg.Dict(x=pg.MISSING_VALUE, y=1)
    assert d.is_partial
    assert 'x' in d.missing_values()
  """

  @property
  def is_partial(self) -> bool:
    """Returns True if this object is partial. Otherwise False.

    An object is considered partial when any of its required fields is missing,
    or at least one member is partial. The subclass can override this method
    to provide a more efficient solution.
    """
    return len(self.missing_values()) > 0  # pylint: disable=g-explicit-length-test

  @abc.abstractmethod
  def missing_values(self, flatten: bool = True) -> Dict[str, Any]:  # pylint: disable=redefined-outer-name
    """Returns missing values from this object.

    Args:
      flatten: If True, convert nested structures into a flattened dict using
        key path (delimited by '.' and '[]') as key.

    Returns:
      A dict of key to MISSING_VALUE.
    """


class Functor(metaclass=abc.ABCMeta):
  """Interface for functor."""

  @abc.abstractmethod
  def __call__(self, *args, **kwargs) -> Any:
    """Calls the functor.

    Args:
      *args: Any positional arguments.
      **kwargs: Any keyword arguments.

    Returns:
      Any value.
    """


def explicit_method_override(method):
  """Decorator that marks a member method as explicitly overridden.

  In PyGlove, many methods are managed by the framework - for example -
  ``pg.Object.__init__``. It's easy for users to override these methods
  unconsciously. Therefore, we introduce this decorator to catch error at
  the first place when such overrides incidentally take place, while allowing
  advanced users to override them.

  Usage::

    class Foo(pg.Object):

      @pg.explicit_method_override
      def __init__(self, *args, **kwargs):
       ...

  Args:
    method: method to explicitly overriden.

  Returns:
    The original method with an explicit overriden stamp.
  """
  setattr(method, '__explicit_override__', True)
  return method


def ensure_explicit_method_override(
    method, error_message: Optional[str] = None) -> None:
  """Returns True if a method is explicitly overridden."""
  if not getattr(method, '__explicit_override__', False):
    if error_message is None:
      error_message = (
          f'{method} is a PyGlove managed method. If you do need to override '
          'it, please decorate the method with `@pg.explicit_method_override`.')
    raise TypeError(error_message)



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic differences."""

from typing import Any, Callable, List, Optional, Tuple, Union

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import base
from pyglove.core.symbolic import list as pg_list
from pyglove.core.symbolic import object as pg_object
from pyglove.core.symbolic.pure_symbolic import PureSymbolic
from pyglove.core.views import html


class Diff(PureSymbolic, pg_object.Object):
  """A value diff between two objects: a 'left' object and a 'right' object.

  If one of them is missing, it may be represented by pg.Diff.MISSING

  For example::

    >>> pg.Diff(3.14, 1.618)
    Diff(left=3.14, right=1.618)
    >>> pg.Diff('hello world', pg.Diff.MISSING)
    Diff(left='hello world', right=MISSING)
  """

  class _Missing:
    """Represents an absent party in a Diff."""

    def __repr__(self):
      return self.__str__()

    def __str__(self):
      return 'MISSING'

    def __eq__(self, other):
      return isinstance(other, Diff._Missing)

    def __ne__(self, other):
      return not self.__eq__(other)

  MISSING = _Missing()

  def _on_bound(self):
    super()._on_bound()
    if self.children:
      if not isinstance(self.left, type):
        raise ValueError(
            f'\'left\' must be a type when \'children\' is specified. '
            f'Encountered: {self.left!r}.')
      if not isinstance(self.right, type):
        raise ValueError(
            f'\'right\' must be a type when \'children\' is specified. '
            f'Encountered: {self.right!r}.')
    self._has_diff = None

  @property
  def is_leaf(self) -> bool:
    """Returns True if current Diff does not contain inner Diff object."""
    return not self.children

  def __bool__(self):
    """Returns True if there is a diff."""
    if self._has_diff is None:
      if base.ne(self.left, self.right):
        has_diff = True
      elif self.children:
        has_diff = any(bool(cd) for cd in self.children.values())
      else:
        has_diff = False
      self._has_diff = has_diff
    return self._has_diff

  def sym_eq(self, other: Any):
    """Override symbolic equality."""
    if super().sym_eq(other):
      return True
    if not bool(self):
      return base.eq(self.left, other)

  @property
  def value(self):
    """Returns the value if left and right are the same."""
    if bool(self):
      raise ValueError(
          f'\'value\' cannot be accessed when \'left\' and \'right\' '
          f'are not the same. Left={self.left!r}, Right={self.right!r}.')
    return self.left

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      root_indent: int = 0,
      **kwargs):
    """Override format to conditionally print the shared value or the diff."""
    if not bool(self):
      if self.value == Diff.MISSING:
        return 'No diff'
      # When there is no diff, but the same value needs to be displayed
      # we simply return the value.
      return object_utils.format(
          self.value, compact, verbose, root_indent, **kwargs)
    if self.is_leaf:
      exclude_keys = kwargs.pop('exclude_keys', None)
      exclude_keys = exclude_keys or set()
      exclude_keys.add('children')
      return super().format(
          compact, verbose, root_indent, exclude_keys=exclude_keys, **kwargs)
    else:
      assert isinstance(self.left, type)
      assert isinstance(self.right, type)
      if self.left is self.right and issubclass(self.left, list):
        return self.children.format(
            compact=compact,
            verbose=verbose,
            root_indent=root_indent,
            cls_name='',
            bracket_type=object_utils.BracketType.SQUARE,
            **kwargs,
        )
      if self.left is self.right:
        cls_name = self.left.__name__
      else:
        cls_name = f'{self.left.__name__}|{self.right.__name__}'
      return self.children.format(
          compact=compact,
          verbose=verbose,
          root_indent=root_indent,
          cls_name=cls_name,
          bracket_type=object_utils.BracketType.ROUND,
          **kwargs,
      )

  # pytype: disable=annotation-type-mismatch
  def _html_tree_view_summary(
      self,
      *,
      view: html.HtmlTreeView,
      title: Optional[str] = None,
      max_summary_len_for_str: int = html.HtmlView.PresetArgValue(40),
      **kwargs,
    ) -> Optional[html.Html]:
  # pytype: enable=annotation-type-mismatch
    if not bool(self):
      v = self.value
      if (isinstance(v, (int, float, bool, type(None)))
          or (isinstance(v, str) and len(v) <= max_summary_len_for_str)):
        return None

      if v == Diff.MISSING:
        return view.summary(
            self,
            title=title or 'Diff',
            max_summary_len_for_str=max_summary_len_for_str,
            **kwargs
        )
      return view.summary(
          self.value,
          title=title,
          max_summary_len_for_str=max_summary_len_for_str,
          **kwargs
      )
    elif self.is_leaf:
      return None
    else:
      assert isinstance(self.left, type), self.left
      assert isinstance(self.right, type), self.right
      if self.left is self.right:
        diff_title = self.left.__name__
      else:
        diff_title = f'{self.left.__name__} | {self.right.__name__}'
      return view.summary(
          self,
          title=title or diff_title,
          max_summary_len_for_str=max_summary_len_for_str,
          **kwargs
      )

  def _html_tree_view_content(
      self,
      *,
      view: html.HtmlTreeView,
      parent: Any,
      root_path: object_utils.KeyPath,
      **kwargs
  ) -> html.Html:
    if not bool(self):
      if self.value == Diff.MISSING:
        root = html.Html.element(
            'span',
            # CSS class already defined in HtmlTreeView.
            css_class=['diff_empty']
        )
      else:
        # When there is no diff, but the same value needs to be displayed
        # we simply return the value.
        root = html.Html.element(
            'div',
            [
                view.content(
                    self.value, parent=parent, root_path=root_path, **kwargs,
                )
            ],
            css_class=['diff_left', 'diff_right']
        )
    elif self.is_leaf:
      root = html.Html.element(
          'div',
          [
              # Render left value.
              lambda: html.Html.element(  # pylint: disable=g-long-ternary
                  'div',
                  [
                      view.render(
                          self.left, root_path=root_path + 'left', **kwargs
                      )
                  ],
                  css_class=['diff_left'],
              ) if self.left != Diff.MISSING else None,
              # Render right value.
              lambda: html.Html.element(  # pylint: disable=g-long-ternary
                  'div',
                  [
                      view.render(
                          self.right, root_path=root_path + 'right', **kwargs
                      )
                  ],
                  css_class=['diff_right'],
              )  if self.right != Diff.MISSING else None,
          ],
          css_class=['diff_value']
      )
    else:
      assert isinstance(self.left, type)
      assert isinstance(self.right, type)

      if issubclass(self.left, list):
        key_fn = int
      else:
        key_fn = lambda k: k

      s = html.Html()
      for k, v in self.children.items():
        k = key_fn(k)
        child_path = root_path + k

        s.write('<tr><td>')
        # Print Key.
        s.write(
            view.object_key(
                k, parent=v, root_path=child_path,
                css_class=None if bool(v) else ['no_diff_key'],
                **kwargs
            ),
        )
        # Print Value.
        s.write(
            '</td><td>',
            view.render(v, root_path=child_path, **kwargs),
            '</td></tr>'
        )
      root = html.Html.element(
          'div',
          [
              '<table>', s, '</table>'
          ],
          css_class=[
              'complex_value', view.css_class_name(self.left)
          ]
      )
    return root

  def _html_style(self) -> List[str]:
    return [
        """
        /* Diff styles. */
        .diff .summary_title::after {
          content: ' (diff)';
          color: #aaa;
        }
        .diff .summary_title {
          background-color: yellow;
        }
        .diff table td {
          padding: 4px;
        }
        .diff_value {
          display: inline-block;
          align-items: center;
        }
        .no_diff_key {
          opacity: 0.4;
        }
        .diff_left.diff_right::after {
          content: '(no diff)';
          margin-left: 0.5em;
          color: #aaa;
          font-style: italic;
        }
        .diff_empty::before {
            content: '(empty)';
            font-style: italic;
            margin-left: 0.5em;
            color: #aaa;
        }
        .diff_value .diff_left::before {
          content: '🇱';
        }
        .diff_value .diff_left > .simple_value {
          background-color: #ffcccc;
        }
        .diff_value .diff_left > details {
          background-color: #ffcccc;
        }
        .diff_value .diff_right::before {
          content: '🇷';
        }
        .diff_value .diff_right > .simple_value {
          background-color: #ccffcc;
        }
        .diff_value .diff_right > details {
          background-color: #ccffcc;
        }
        """
    ]


# NOTE(daiyip): we add the symbolic attribute to Diff after its declaration
# since we need to access Diff.MISSING as the default value for `left` and
# `right`.
pg_object.members([
    ('left', pg_typing.Any(default=Diff.MISSING),
     'The left-hand object being compared.'),
    ('right', pg_typing.Any(default=Diff.MISSING),
     'The right-hand object being compared.'),
    ('children', pg_typing.Dict([
        (pg_typing.StrKey(), pg_typing.Object(Diff), 'Child node.')
    ]))
])(Diff)


def diff(
    left: Any,
    right: Any,
    flatten: bool = False,
    collapse: Union[bool, str, Callable[[Any, Any], bool]] = 'same_type',
    mode: str = 'diff') -> object_utils.Nestable[Diff]:
  """Inspect the symbolic diff between two objects.

  For example::

    @pg.members([
      ('x', pg.Any()),
      ('y', pg.Any())
    ])
    class A(pg.Object):
      pass

    @pg.members([
      ('z', pg.Any().noneable())
    ])
    class B(A):
      pass


    # Diff the same object.
    pg.diff(A(1, 2), A(1, 2))

    >> No diff

    # Diff the same object with mode 'same'.
    pg.diff(A(1, 2), A(1, 2), mode='same')

    >> A(1, 2)

    # Diff different objects of the same type.
    pg.diff(A(1, 2), A(1, 3))

    >> A(
    >>   y = Diff(
    >>     left=2,
    >>     right=3
    >>   )
    >>  )

    # Diff objects of different type.
    pg.diff(A(1, 2), B(1, 3))

    >> Diff(
    >>    left = A(
    >>      x = 1,
    >>      y = 2
    >>    ),
    >>    right = B(
    >>      x = 1,
    >>      y = 3,
    >>      z = None
    >>    )

    # Diff objects of different type with collapse.
    pg.diff(A(1, 2), B(1, 3), collapse=True)

    >> A|B (
    >>   y = Diff(
    >>     left = 2,
    >>     right = 3,
    >>   ),
    >>   z = Diff(
    >>     left = MISSING,
    >>     right = None
    >>   )
    >> )

    # Diff objects of different type with collapse and flatten.
    # Object type is included in key '_type'.
    pg.diff(A(1, pg.Dict(a=1)), B(1, pg.Dict(a=2)), collapse=True, flatten=True)

    >> {
    >>    'y.a': Diff(1, 2),
    >>    'z', Diff(MISSING, None),
    >>    '_type': Diff(A, B)
    >> }

  Args:
    left: The left object to compare.
    right: The right object to compare.
    flatten: If True, returns a level-1 dict with diff keys flattened. Otherwise
      preserve the hierarchy of the diff result.
    collapse: One of a boolean value, string or a callable object that indicates
      whether to collapse two different values. The default value 'same_type'
      means only collapse when the two values are of the same type.
    mode: Diff mode, should be one of ['diff', 'same', 'both']. For 'diff' mode
      (the default), the return value contains only different values. For 'same'
      mode, the return value contains only same values. For 'both', the return
      value contains both different and same values.

  Returns:
    A `Diff` object when flatten is False. Otherwise a dict of string (key path)
    to `Diff`.
  """
  def _should_collapse(left, right):
    if isinstance(left, dict):
      if isinstance(right, dict):
        return True
    elif isinstance(left, list):
      return isinstance(right, list)

    if (isinstance(left, (dict, base.Symbolic))
        and isinstance(right, (dict, base.Symbolic))):
      if collapse == 'same_type':
        return type(left) is type(right)
      elif callable(collapse):
        return collapse(left, right)
      elif isinstance(collapse, bool):
        return collapse
      else:
        raise ValueError(f'Unsupported `collapse` value: {collapse!r}')
    else:
      return False

  def _add_child_diff(diff_container, key, value, child_has_diff):
    if ((mode != 'same' and child_has_diff)
        or (mode != 'diff' and not child_has_diff)):
      diff_container[key] = value

  def _get_container_ops(container):
    if isinstance(container, dict):
      return container.__contains__, container.__getitem__, container.items
    else:
      assert isinstance(container, base.Symbolic)
      return container.sym_hasattr, container.sym_getattr, container.sym_items

  def _diff(x, y) -> Tuple[object_utils.Nestable[Diff], bool]:
    if x is y or x == y:
      return (Diff(x, y), False)
    if not _should_collapse(x, y):
      return (Diff(x, y), True)

    diff_value, has_diff = {}, False
    if isinstance(x, list):
      assert isinstance(y, list)
      def _child(l, index):
        return l[i] if index < len(l) else Diff.MISSING
      for i in range(max(len(x), len(y))):
        child_diff, child_has_diff = _diff(_child(x, i), _child(y, i))
        has_diff = has_diff or child_has_diff
        _add_child_diff(diff_value, str(i), child_diff, child_has_diff)
      diff_value = Diff(pg_list.List, pg_list.List, children=diff_value)
    else:
      assert isinstance(x, (dict, base.Symbolic))
      assert isinstance(y, (dict, base.Symbolic))

      x_haskey, _, x_items = _get_container_ops(x)
      y_haskey, y_getitem, y_items = _get_container_ops(y)

      for k, xv in x_items():
        yv = y_getitem(k) if y_haskey(k) else Diff.MISSING
        child_diff, child_has_diff = _diff(xv, yv)
        has_diff = has_diff or child_has_diff
        _add_child_diff(diff_value, k, child_diff, child_has_diff)

      for k, yv in y_items():
        if not x_haskey(k):
          child_diff, _ = _diff(Diff.MISSING, yv)
          has_diff = True
          _add_child_diff(diff_value, k, child_diff, True)

      xt, yt = type(x), type(y)
      same_type = xt is yt
      if not same_type:
        has_diff = True

      if flatten:
        # Put type difference with key '_type'. Since symbolic
        # fields will not start with underscore, so there should be
        # no clash.
        if not same_type or mode != 'diff':
          diff_value['_type'] = Diff(xt, yt)
      else:
        diff_value = Diff(xt, yt, children=diff_value)
    return diff_value, has_diff

  diff_value, has_diff = _diff(left, right)
  if not has_diff and mode == 'diff':
    diff_value = Diff()
  if flatten:
    diff_value = object_utils.flatten(diff_value)
  return diff_value



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Any, List, Sequence
import unittest

from pyglove.core.views.html import base
from pyglove.core.views.html import tree_view

Html = base.Html
KeyPath = tree_view.KeyPath
KeyPathSet = tree_view.KeyPathSet


class TestCase(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self._view = tree_view.HtmlTreeView()

  def assert_style(self, html, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = html.style_section.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual.strip(), expected)

  def assert_content(self, html, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = html.content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual.strip(), expected)

  def assert_count(self, html, part, expected):
    content = html.content.strip()
    actual = content.count(part)
    if actual != expected:
      print(content)
    self.assertEqual(actual, expected)


class TooltipTest(TestCase):

  def test_style(self):
    self.assert_style(
        self._view.tooltip(
            'This <hello><world></world></hello>.',
            name='name',
        ),
        """
        <style>
        /* Tooltip styles. */
        span.tooltip {
          visibility: hidden;
          white-space: pre-wrap;
          font-weight: normal;
          background-color: #484848;
          color: #fff;
          padding: 10px;
          border-radius: 6px;
          position: absolute;
          z-index: 1;
        }
        </style>
        """
    )

  def test_render(self):
    self.assert_content(
        self._view.tooltip(
            'This <hello><world></world></hello>.',
            name='name',
        ),
        """
        <span class="tooltip str">&#x27;This &lt;hello&gt;&lt;world&gt;&lt;/world&gt;&lt;/hello&gt;.&#x27;</span>
        """
    )
    self.assert_content(
        self._view.tooltip(
            1,
            content=Html.element('div', ['hello']),
            name='name',
        ),
        """
        <span class="tooltip int"><div>hello</div></span>
        """
    )


class SummaryTest(TestCase):

  def test_style(self):
    self.assert_style(
        self._view.summary(1, enable_summary=True),
        """
        <style>
        /* Tooltip styles. */
        span.tooltip {
          visibility: hidden;
          white-space: pre-wrap;
          font-weight: normal;
          background-color: #484848;
          color: #fff;
          padding: 10px;
          border-radius: 6px;
          position: absolute;
          z-index: 1;
        }
        /* Summary styles. */
        details.pyglove summary {
          font-weight: bold;
          margin: -0.5em -0.5em 0;
          padding: 0.5em;
        }
        .summary_name {
          display: inline;
          padding: 0 5px;
        }
        .summary_title {
          display: inline;
        }
        .summary_name + div.summary_title {
          display: inline;
          color: #aaa;
        }
        .summary_title:hover + span.tooltip {
          visibility: visible;
        }
        /* Type-specific styles. */
        .pyglove.str .summary_title {
          color: darkred;
          font-style: italic;
        }
        </style>
        """
    )

  def test_enable_summary(self):
    self.assertIsNone(
        self._view.summary(1, enable_summary=None),
    )
    self.assertIsNone(
        self._view.summary('foo', enable_summary=False),
    )
    self.assertIsNone(
        self._view.summary('foo', enable_summary=None),
    )
    self.assert_content(
        self._view.summary('foo', enable_summary=True),
        """
        <summary><div class="summary_title">&#x27;foo&#x27;</div><span class="tooltip str">&#x27;foo&#x27;</span></summary>
        """
    )
    self.assert_content(
        self._view.summary(
            'foo', enable_summary=None, max_summary_len_for_str=1
        ),
        """
        <summary><div class="summary_title">&#x27;f...&#x27;</div><span class="tooltip str">&#x27;foo&#x27;</span></summary>
        """
    )

  def test_max_summary_len_for_str(self):
    self.assert_content(
        self._view.summary(
            'abcdefg',
            max_summary_len_for_str=5
        ),
        """
        <summary><div class="summary_title">&#x27;abcde...&#x27;</div><span class="tooltip str">&#x27;abcdefg&#x27;</span></summary>
        """
    )

  def test_render_with_name(self):
    self.assert_content(
        self._view.summary('foo', name='x'),
        """
        <summary><div class="summary_name">x</div><div class="summary_title">&#x27;foo&#x27;</div><span class="tooltip str">&#x27;foo&#x27;</span></summary>
        """
    )

  def test_enable_summary_tooltip(self):
    self.assert_content(
        self._view.summary(
            'foo', enable_summary=True, enable_summary_tooltip=True
        ),
        """
        <summary><div class="summary_title">&#x27;foo&#x27;</div><span class="tooltip str">&#x27;foo&#x27;</span></summary>
        """
    )
    self.assert_content(
        self._view.summary(
            'foo', enable_summary=True, enable_summary_tooltip=False
        ),
        """
        <summary><div class="summary_title">&#x27;foo&#x27;</div></summary>
        """
    )


class ContentTest(TestCase):

  def test_style(self):
    self.assert_style(
        self._view.content(dict(x=1)),
        """
        <style>
        /* Tooltip styles. */
        span.tooltip {
          visibility: hidden;
          white-space: pre-wrap;
          font-weight: normal;
          background-color: #484848;
          color: #fff;
          padding: 10px;
          border-radius: 6px;
          position: absolute;
          z-index: 1;
        }
        /* Object key styles. */
        .object_key {
          margin-right: 0.25em;
        }
        .object_key:hover + .tooltip {
          visibility: visible;
          background-color: darkblue;
        }
        .object_key.str {
          color: gray;
          border: 1px solid lightgray;
          background-color: ButtonFace;
          border-radius: 0.2em;
          padding: 0.3em;
        }
        .object_key.int::before{
          content: '[';
        }
        .object_key.int::after{
          content: ']';
        }
        .object_key.int{
          border: 0;
          color: lightgray;
          background-color: transparent;
          border-radius: 0;
          padding: 0;
        }
        /* Simple value styles. */
        .simple_value {
          color: blue;
          display: inline-block;
          white-space: pre-wrap;
          padding: 0.2em;
          margin-top: 0.15em;
        }
        .simple_value.str {
          color: darkred;
          font-style: italic;
        }
        .simple_value.int, .simple_value.float {
          color: darkblue;
        }
        /* Complex value styles. */
        span.empty_container::before {
            content: '(empty)';
            font-style: italic;
            margin-left: 0.5em;
            color: #aaa;
        }
        </style>
        """
    )

  def test_simple_types(self):
    self.assert_content(
        self._view.content(1),
        """
        <span class="simple_value int">1</span>
        """
    )
    self.assert_content(
        self._view.content(1.5),
        """
        <span class="simple_value float">1.5</span>
        """
    )
    self.assert_content(
        self._view.content(True),
        """
        <span class="simple_value bool">True</span>
        """
    )
    self.assert_content(
        self._view.content(None),
        """
        <span class="simple_value none-type">None</span>
        """
    )
    self.assert_content(
        self._view.content('<foo>'),
        """
        <span class="simple_value str">&#x27;&lt;foo&gt;&#x27;</span>
        """
    )
    self.assert_content(
        self._view.content(
            '<hello><world> \nto everyone.', max_summary_len_for_str=10
        ),
        """
        <span class="simple_value str">&lt;hello&gt;&lt;world&gt; 
        to everyone.</span>
        """
    )

  def test_list(self):
    self.assert_content(
        self._view.content([1, 2, 'abc']),
        """
        <div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[1]</span></td><td><div><span class="simple_value int">2</span></div></td></tr><tr><td><span class="object_key int">2</span><span class="tooltip key-path">[2]</span></td><td><div><span class="simple_value str">&#x27;abc&#x27;</span></div></td></tr></table></div>
        """
    )
    self.assert_content(
        self._view.content([]),
        """
        <div class="complex_value list"><span class="empty_container"></span></div>
        """
    )

  def test_tuple(self):
    self.assert_content(
        self._view.content((1, True)),
        """
        <div class="complex_value tuple"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[1]</span></td><td><div><span class="simple_value bool">True</span></div></td></tr></table></div>
        """
    )
    self.assert_content(
        self._view.content(()),
        """
        <div class="complex_value tuple"><span class="empty_container"></span></div>
        """
    )

  def test_dict(self):
    self.assert_content(
        self._view.content(dict(x=1, y='foo')),
        """
        <div class="complex_value dict"><table><tr><td><span class="object_key str">x</span><span class="tooltip key-path">x</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">y</span><span class="tooltip key-path">y</span></td><td><div><span class="simple_value str">&#x27;foo&#x27;</span></div></td></tr></table></div>
        """
    )
    self.assert_content(
        self._view.content({}),
        """
        <div class="complex_value dict"><span class="empty_container"></span></div>
        """
    )

  def test_custom_types(self):

    class Foo:
      def __str__(self):
        return '<Foo></Foo>'

    self.assert_content(
        self._view.content(Foo()),
        """
        <span class="simple_value foo">&lt;Foo&gt;&lt;/Foo&gt;</span>
        """
    )

  def test_custom_key_value_render(self):
    def render_key(key, **kwargs):
      del kwargs
      return Html.element('span', [f'custom {key}'])

    def render_value(value, **kwargs):
      del kwargs
      return Html.element('span', [f'custom {value}'])

    self.assert_content(
        self._view.complex_value(
            dict(x=1, y='foo'),
            parent=None,
            root_path=KeyPath(),
            render_key_fn=render_key,
            render_value_fn=render_value,
        ),
        """
        <div class="complex_value none-type"><table><tr><td><span>custom x</span></td><td><div><span>custom 1</span></div></td></tr><tr><td><span>custom y</span></td><td><div><span>custom foo</span></div></td></tr></table></div>
        """
    )

  def test_nesting(self):

    class Foo:
      def __str__(self):
        return '<Foo></Foo>'

    self.assert_content(
        self._view.content(
            [
                dict(
                    x=[(1, 2)],
                    y=['b', Foo()],
                ),
                1,
                [1, dict(xx=1, yy='a')]
            ]
        ),
        """
        <div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0]</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div><span class="tooltip dict">{
          &#x27;x&#x27;: [(1, 2)],
          &#x27;y&#x27;: [&#x27;b&#x27;, &lt;Foo&gt;&lt;/Foo&gt;]
        }</span></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">x</span><span class="tooltip key-path">[0].x</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[(1, 2)]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].x[0]</span></td><td><div><details class="pyglove tuple"><summary><div class="summary_title">tuple(...)</div><span class="tooltip tuple">(1, 2)</span></summary><div class="complex_value tuple"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].x[0][0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[0].x[0][1]</span></td><td><div><span class="simple_value int">2</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key str">y</span><span class="tooltip key-path">[0].y</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[&#x27;b&#x27;, &lt;Foo&gt;&lt;/Foo&gt;]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].y[0]</span></td><td><div><span class="simple_value str">&#x27;b&#x27;</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[0].y[1]</span></td><td><div><details class="pyglove foo"><summary><div class="summary_title">Foo(...)</div><span class="tooltip foo">&lt;Foo&gt;&lt;/Foo&gt;</span></summary><span class="simple_value foo">&lt;Foo&gt;&lt;/Foo&gt;</span></details></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[1]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">2</span><span class="tooltip key-path">[2]</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[1, {
            &#x27;xx&#x27;: 1,
            &#x27;yy&#x27;: &#x27;a&#x27;
          }]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[2][0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[2][1]</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div><span class="tooltip dict">{
          &#x27;xx&#x27;: 1,
          &#x27;yy&#x27;: &#x27;a&#x27;
        }</span></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">xx</span><span class="tooltip key-path">[2][1].xx</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">yy</span><span class="tooltip key-path">[2][1].yy</span></td><td><div><span class="simple_value str">&#x27;a&#x27;</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr></table></div>
        """
    )

  def test_enable_disable_tooltip(self):

    class Foo:
      def __str__(self):
        return '<Foo></Foo>'

    value = [
        dict(
            x=[(1, 2)],
            y=['b', Foo()],
        ),
        1,
        [1, dict(xx=1, yy='a')]
    ]
    self.assert_content(
        self._view.content(value, enable_summary_tooltip=False),
        """
        <div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0]</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">x</span><span class="tooltip key-path">[0].x</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].x[0]</span></td><td><div><details class="pyglove tuple"><summary><div class="summary_title">tuple(...)</div></summary><div class="complex_value tuple"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].x[0][0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[0].x[0][1]</span></td><td><div><span class="simple_value int">2</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key str">y</span><span class="tooltip key-path">[0].y</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].y[0]</span></td><td><div><span class="simple_value str">&#x27;b&#x27;</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[0].y[1]</span></td><td><div><details class="pyglove foo"><summary><div class="summary_title">Foo(...)</div></summary><span class="simple_value foo">&lt;Foo&gt;&lt;/Foo&gt;</span></details></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[1]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">2</span><span class="tooltip key-path">[2]</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[2][0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[2][1]</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">xx</span><span class="tooltip key-path">[2][1].xx</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">yy</span><span class="tooltip key-path">[2][1].yy</span></td><td><div><span class="simple_value str">&#x27;a&#x27;</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr></table></div>
        """
    )
    self.assert_content(
        self._view.content(value, enable_key_tooltip=False),
        """
        <div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div><span class="tooltip dict">{
          &#x27;x&#x27;: [(1, 2)],
          &#x27;y&#x27;: [&#x27;b&#x27;, &lt;Foo&gt;&lt;/Foo&gt;]
        }</span></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">x</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[(1, 2)]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><details class="pyglove tuple"><summary><div class="summary_title">tuple(...)</div><span class="tooltip tuple">(1, 2)</span></summary><div class="complex_value tuple"><table><tr><td><span class="object_key int">0</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><span class="simple_value int">2</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key str">y</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[&#x27;b&#x27;, &lt;Foo&gt;&lt;/Foo&gt;]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><span class="simple_value str">&#x27;b&#x27;</span></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><details class="pyglove foo"><summary><div class="summary_title">Foo(...)</div><span class="tooltip foo">&lt;Foo&gt;&lt;/Foo&gt;</span></summary><span class="simple_value foo">&lt;Foo&gt;&lt;/Foo&gt;</span></details></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">2</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[1, {
            &#x27;xx&#x27;: 1,
            &#x27;yy&#x27;: &#x27;a&#x27;
          }]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div><span class="tooltip dict">{
          &#x27;xx&#x27;: 1,
          &#x27;yy&#x27;: &#x27;a&#x27;
        }</span></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">xx</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">yy</span></td><td><div><span class="simple_value str">&#x27;a&#x27;</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr></table></div>
        """
    )
    self.assert_content(
        self._view.content(
            value, enable_summary_tooltip=False, enable_key_tooltip=False
        ),
        """
        <div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">x</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><details class="pyglove tuple"><summary><div class="summary_title">tuple(...)</div></summary><div class="complex_value tuple"><table><tr><td><span class="object_key int">0</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><span class="simple_value int">2</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key str">y</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><span class="simple_value str">&#x27;b&#x27;</span></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><details class="pyglove foo"><summary><div class="summary_title">Foo(...)</div></summary><span class="simple_value foo">&lt;Foo&gt;&lt;/Foo&gt;</span></details></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">2</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">xx</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">yy</span></td><td><div><span class="simple_value str">&#x27;a&#x27;</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr></table></div>
        """
    )

  def test_include_exclude_immediate_child_keys(self):
    self.assert_count(
        self._view.content(
            dict(a='x', b=dict(e='y'), c='z', d='w', e='v', f='u'),
            include_keys=['a', 'b', 'c', 'd'],
            exclude_keys=['c', 'e'],
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        'object_key',
        3 + 1  # 'a', 'b', 'd' at the first level and 'e' at the second level.
    )

  def test_special_keys(self):
    self.assert_content(
        self._view.content(
            dict(a='x', b=dict(e='y'), c='z', d='w', e='v', f='u'),
            special_keys=['c', 'b'],
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <div class="complex_value dict"><div class="special_value"><details class="pyglove str"><summary><div class="summary_name">c</div><div class="summary_title">&#x27;z&#x27;</div></summary><span class="simple_value str">&#x27;z&#x27;</span></details></div><div class="special_value"><details class="pyglove dict"><summary><div class="summary_name">b</div><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">e</span></td><td><div><span class="simple_value str">&#x27;y&#x27;</span></div></td></tr></table></div></details></div><table><tr><td><span class="object_key str">a</span></td><td><div><span class="simple_value str">&#x27;x&#x27;</span></div></td></tr><tr><td><span class="object_key str">d</span></td><td><div><span class="simple_value str">&#x27;w&#x27;</span></div></td></tr><tr><td><span class="object_key str">e</span></td><td><div><span class="simple_value str">&#x27;v&#x27;</span></div></td></tr><tr><td><span class="object_key str">f</span></td><td><div><span class="simple_value str">&#x27;u&#x27;</span></div></td></tr></table></div>
        """
    )
    self.assert_count(
        self._view.content(
            dict(a='x', b=dict(e='y'), c='z', d='w', e='v', f='u'),
            special_keys=['b', 'd'],
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        'special_value',
        2
    )

  def test_filter(self):
    self.assert_content(
        self._view.content(
            dict(a='x', b=dict(e='y'), c='z', d='w', e='v', f='u'),
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            filter=(lambda k, v, p: len(k) > 1 or isinstance(v, dict))
        ),
        """
        <div class="complex_value dict"><table><tr><td><span class="object_key str">b</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">e</span></td><td><div><span class="simple_value str">&#x27;y&#x27;</span></div></td></tr></table></div></details></div></td></tr></table></div>
        """
    )

  def test_highlight_lowlight(self):
    self.assert_content(
        self._view.content(
            dict(a=1, b=dict(e='y'), c=2, d='w', e=3, f='u'),
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            highlight_keys=(lambda k, v, p: isinstance(v, int)),
            lowlight_keys=(lambda k, v, p: isinstance(v, str)),
        ),
        """
        <div class="complex_value dict"><table><tr><td><span class="object_key str">a</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">b</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">e</span></td><td><div><span class="simple_value str">&#x27;y&#x27;</span></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key str">c</span></td><td><div><span class="simple_value int">2</span></div></td></tr><tr><td><span class="object_key str">d</span></td><td><div><span class="simple_value str">&#x27;w&#x27;</span></div></td></tr><tr><td><span class="object_key str">e</span></td><td><div><span class="simple_value int">3</span></div></td></tr><tr><td><span class="object_key str">f</span></td><td><div><span class="simple_value str">&#x27;u&#x27;</span></div></td></tr></table></div>
        """
    )

  def test_collapse_level(self):
    x = dict(
        a=dict(
            b=dict(
                c=dict(
                    d=dict()
                )
            )
        ),
        aa=dict(
            bb=dict(
                cc=dict(
                    dd=dict()
                )
            )
        )
    )
    self.assert_count(
        self._view.content(x, collapse_level=0),
        'open',
        0
    )
    # There is no summary section, so there is no root.
    self.assert_count(
        self._view.content(x, collapse_level=1),
        'open',
        0
    )
    self.assert_count(
        self._view.content(x, collapse_level=2),
        'open',
        2
    )
    self.assert_count(
        self._view.content(x, collapse_level=3),
        'open',
        4
    )
    self.assert_count(
        self._view.content(x, collapse_level=4),
        'open',
        6
    )
    self.assert_count(
        self._view.content(x, collapse_level=None),
        'open',
        8
    )

  def test_uncollapse(self):
    x = dict(
        a=dict(
            b=dict(
                c=dict(
                    d=dict()
                )
            )
        ),
        aa=dict(
            bb=dict(
                cc=dict(
                    dd=dict()
                )
            )
        )
    )
    self.assert_count(
        self._view.content(x, uncollapse=['a.b.c']),
        'open',
        3
    )
    self.assert_count(
        self._view.content(x, uncollapse=['aa.bb.cc.dd', 'a.b']),
        'open',
        6
    )
    # Use both collapse_level and uncollapse.
    self.assert_count(
        self._view.content(x, collapse_level=2, uncollapse=['aa.bb.cc']),
        'open',
        4
    )
    self.assert_count(
        self._view.content(
            x, collapse_level=0, uncollapse=lambda k, v, p: len(k) < 3
        ),
        'open',
        4
    )


class RenderTest(TestCase):

  def test_render(self):
    class Foo:
      def __str__(self):
        return '<Foo></Foo>'

    self.assert_content(
        self._view.render(
            [
                dict(
                    x=[(1, 2)],
                    y=['b', Foo()],
                ),
                1,
                [1, dict(xx=1, yy='a')]
            ],
        ),
        """
        <details open class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[
          {
            &#x27;x&#x27;: [(1, 2)],
            &#x27;y&#x27;: [&#x27;b&#x27;, &lt;Foo&gt;&lt;/Foo&gt;]
          },
          1,
          [1, {
              &#x27;xx&#x27;: 1,
              &#x27;yy&#x27;: &#x27;a&#x27;
            }]
        ]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0]</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div><span class="tooltip dict">{
          &#x27;x&#x27;: [(1, 2)],
          &#x27;y&#x27;: [&#x27;b&#x27;, &lt;Foo&gt;&lt;/Foo&gt;]
        }</span></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">x</span><span class="tooltip key-path">[0].x</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[(1, 2)]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].x[0]</span></td><td><div><details class="pyglove tuple"><summary><div class="summary_title">tuple(...)</div><span class="tooltip tuple">(1, 2)</span></summary><div class="complex_value tuple"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].x[0][0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[0].x[0][1]</span></td><td><div><span class="simple_value int">2</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key str">y</span><span class="tooltip key-path">[0].y</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[&#x27;b&#x27;, &lt;Foo&gt;&lt;/Foo&gt;]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[0].y[0]</span></td><td><div><span class="simple_value str">&#x27;b&#x27;</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[0].y[1]</span></td><td><div><details class="pyglove foo"><summary><div class="summary_title">Foo(...)</div><span class="tooltip foo">&lt;Foo&gt;&lt;/Foo&gt;</span></summary><span class="simple_value foo">&lt;Foo&gt;&lt;/Foo&gt;</span></details></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[1]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">2</span><span class="tooltip key-path">[2]</span></td><td><div><details class="pyglove list"><summary><div class="summary_title">list(...)</div><span class="tooltip list">[1, {
            &#x27;xx&#x27;: 1,
            &#x27;yy&#x27;: &#x27;a&#x27;
          }]</span></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span><span class="tooltip key-path">[2][0]</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key int">1</span><span class="tooltip key-path">[2][1]</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div><span class="tooltip dict">{
          &#x27;xx&#x27;: 1,
          &#x27;yy&#x27;: &#x27;a&#x27;
        }</span></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">xx</span><span class="tooltip key-path">[2][1].xx</span></td><td><div><span class="simple_value int">1</span></div></td></tr><tr><td><span class="object_key str">yy</span><span class="tooltip key-path">[2][1].yy</span></td><td><div><span class="simple_value str">&#x27;a&#x27;</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr></table></div></details>
        """
    )


class ExtensionTest(TestCase):

  def test_no_overrides(self):

    class Foo(tree_view.HtmlTreeView.Extension):

      def __str__(self):
        return 'Foo()'

    self.assert_content(
        self._view.summary(
            Foo(),
            enable_summary=True,
            max_summary_len_for_str=10,
            enable_tooltip=True,
        ),
        """
        <summary><div class="summary_title">Foo(...)</div><span class="tooltip foo">Foo()</span></summary>
        """
    )
    self.assert_content(
        self._view.content(
            Foo(),
        ),
        """
        <span class="simple_value foo">Foo()</span>
        """
    )

  def test_options_overrides(self):

    class Foo(tree_view.HtmlTreeView.Extension):

      def __str__(self):
        return 'Foo()'

      def _html_style(self) -> List[str]:
        return [
            """
            /* Foo style */
            """
        ]

      def _html_element_class(self) -> List[str]:
        return ['foo', 'bar']

      def _html_tree_view_special_keys(self) -> Sequence[str]:
        return ['x']

      def _html_tree_view_include_keys(self) -> Sequence[str]:
        return ['x', 'y', 'z', 'w']

      def _html_tree_view_exclude_keys(self) -> Sequence[str]:
        return ['z']

      def _html_tree_view_collapse_level(self) -> int:
        return 1

      def _html_tree_view_uncollapse(self) -> KeyPathSet:
        return KeyPathSet(['x.a', 'y.b'], include_intermediate=True)

      def _html_tree_view_content(
          self,
          *,
          view: tree_view.HtmlTreeView,
          parent: Any,
          root_path: KeyPath,
          **kwargs) -> base.Html:
        return view.complex_value(
            kv=dict(
                x=dict(a=dict(foo=2)),
                y=dict(b=dict(bar=3)),
                z=dict(c=dict(baz=4)),
                w=dict(d=dict(qux=5)),
            ),
            parent=self,
            root_path=root_path,
            **kwargs
        )

    self.assertIn(
        '/* Foo style */',
        self._view.render(
            Foo(),
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            collapse_level=0,
        ).styles.content,
    )
    self.assert_content(
        self._view.render(
            [Foo()],
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            collapse_level=0,
        ),
        """
        <details class="pyglove list"><summary><div class="summary_title">list(...)</div></summary><div class="complex_value list"><table><tr><td><span class="object_key int">0</span></td><td><div><details open class="pyglove foo bar"><summary><div class="summary_title">Foo(...)</div></summary><div class="complex_value foo bar"><div class="special_value"><details open class="pyglove dict"><summary><div class="summary_name">x</div><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">a</span></td><td><div><details open class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">foo</span></td><td><div><span class="simple_value int">2</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div><table><tr><td><span class="object_key str">y</span></td><td><div><details open class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">b</span></td><td><div><details open class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">bar</span></td><td><div><span class="simple_value int">3</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr><tr><td><span class="object_key str">w</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">d</span></td><td><div><details class="pyglove dict"><summary><div class="summary_title">dict(...)</div></summary><div class="complex_value dict"><table><tr><td><span class="object_key str">qux</span></td><td><div><span class="simple_value int">5</span></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr></table></div></details></div></td></tr></table></div></details>
        """
    )

  def test_behavior_overrides(self):

    class Foo(tree_view.HtmlTreeView.Extension):

      def _html_tree_view_summary(
          self,
          *,
          view: tree_view.HtmlTreeView,
          **kwargs
      ):
        kwargs.pop('title', None)
        return view.summary(
            self,
            title='MyFoo',
            **kwargs
        )

      def _html_tree_view_content(
          self,
          **kwargs
      ):
        del kwargs
        return 'Content of MyFoo'

      def _html_tree_view_tooltip(
          self,
          *,
          view: tree_view.HtmlTreeView,
          **kwargs
      ):
        del kwargs
        return view.tooltip(
            value=self,
            content='Tooltip <MyFoo>'
        )

    self.assert_content(
        self._view.render(
            Foo(),
        ),
        """
        <details open class="pyglove foo"><summary><div class="summary_title">MyFoo</div><span class="tooltip foo">Tooltip <MyFoo></span></summary>Content of MyFoo</details>
        """
    )
    self.assert_content(
        self._view.render(
            Foo(),
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove foo"><summary><div class="summary_title">MyFoo</div></summary>Content of MyFoo</details>
        """
    )
    self.assert_content(
        self._view.render(
            Foo(),
            enable_summary=False,
        ),
        """
        Content of MyFoo
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for profiling."""

import dataclasses
import time
from typing import Dict, List, Optional

from pyglove.core.object_utils import thread_local


class TimeIt:
  """Context manager for timing the execution of a code block."""

  @dataclasses.dataclass(frozen=True)
  class Status:
    """Status of a single `pg.timeit`."""
    name: str
    elapse: float
    has_ended: bool
    error: Optional[Exception]

    @property
    def has_started(self) -> bool:
      """Returns whether the context has started."""
      return self.elapse > 0

    @property
    def has_error(self) -> bool:
      """Returns whether the context has error."""
      return self.error is not None

  def __init__(self, name: str):
    self._name = name
    self._start_time = None
    self._end_time = None
    self._child_contexts = {}
    self._error = None
    self._parent = None

  @property
  def name(self) -> str:
    """Returns the name of the context."""
    return self._name

  @property
  def children(self) -> List['TimeIt']:
    """Returns child contexts."""
    return list(self._child_contexts.values())

  def add(self, context: 'TimeIt'):
    """Adds a child context."""
    if context.name in self._child_contexts:
      raise ValueError(f'`timeit` with name {context.name!r} already exists.')
    self._child_contexts[context.name] = context

  def start(self):
    """Starts timing."""
    self._start_time = time.time()

  def end(self, error: Optional[BaseException] = None) -> bool:
    """Ends timing."""
    if not self.has_ended:
      self._end_time = time.time()
      self._error = error
      return True
    return False

  @property
  def has_started(self) -> bool:
    """Returns whether the context has started."""
    return self._start_time is not None

  @property
  def has_ended(self) -> bool:
    """Returns whether the context has ended."""
    return self._end_time is not None

  @property
  def start_time(self) -> Optional[float]:
    """Returns start time."""
    return self._start_time

  @property
  def end_time(self) -> Optional[float]:
    """Returns end time."""
    return self._end_time

  @property
  def error(self) -> Optional[BaseException]:
    """Returns error."""
    return self._error

  @property
  def has_error(self) -> bool:
    """Returns whether the context has error."""
    return self._error is not None

  @property
  def elapse(self) -> float:
    """Returns the elapse since start until end."""
    if self._start_time is None:
      return 0
    if self._end_time is None:
      return time.time() - self._start_time
    return self._end_time - self._start_time  # pytype: disable=unsupported-operands

  def status(self) -> Dict[str, Status]:
    """Gets the status of all `timeit` under this context."""
    result = {
        self.name: TimeIt.Status(
            name=self.name, elapse=self.elapse,
            has_ended=self.has_ended, error=self._error,
        )
    }
    for child in self._child_contexts.values():
      child_result = child.status()
      for k, v in child_result.items():
        result[f'{self.name}.{k}'] = v
    return result

  def __enter__(self):
    parent = thread_local.thread_local_get('__timing_context__', None)
    if parent is not None:
      parent.add(self)
      self._parent = parent
    thread_local.thread_local_set('__timing_context__', self)
    self.start()
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    del exc_type, traceback
    self.end(exc_value)
    if self._parent is None:
      thread_local.thread_local_del('__timing_context__')
    else:
      thread_local.thread_local_set('__timing_context__', self._parent)


def timeit(name: str) -> TimeIt:
  """Context manager to time a block of code."""
  return TimeIt(name)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""HTML Tree View (The default view for PyGlove objects)."""

import inspect
import re
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Union

from pyglove.core import object_utils
from pyglove.core.views.html import base


KeyPath = object_utils.KeyPath
KeyPathSet = object_utils.KeyPathSet
Html = base.Html
HtmlView = base.HtmlView


# pytype: disable=annotation-type-mismatch


class HtmlTreeView(HtmlView):
  """HTML Tree View."""

  VIEW_ID = 'html-tree-view'

  class Extension(HtmlView.Extension):
    """The base class for extensions for HtmlTreeView."""

    #
    # Default extension-level rendering options overrides.
    #

    def _html_tree_view_special_keys(self) -> Sequence[str]:
      """Returns the special keys to display (at the immediate child level)."""
      return []

    def _html_tree_view_include_keys(self) -> Optional[Sequence[str]]:
      """Returns the keys to include (at the immediate child level)."""
      return None

    def _html_tree_view_exclude_keys(self) -> Sequence[str]:
      """Returns the keys to include (at the immediate child level)."""
      return []

    def _html_tree_view_collapse_level(self) -> int:
      """Returns the level to collapse the tree (relative to this node)."""
      return 0

    def _html_tree_view_uncollapse(self) -> KeyPathSet:
      """Returns the node paths (relative to current node) to uncollapse."""
      return KeyPathSet()

    #
    # Default behavior overrides.
    #

    def _html_tree_view_render(
        self,
        *,
        view: 'HtmlView',
        name: Optional[str],
        parent: Any,
        root_path: KeyPath,
        title: Union[str, Html, None] = None,
        special_keys: Optional[Sequence[Union[int, str]]] = None,
        include_keys: Optional[Iterable[Union[int, str]]] = None,
        exclude_keys: Optional[Iterable[Union[int, str]]] = None,
        collapse_level: Optional[int] = 1,
        uncollapse: Union[KeyPathSet, base.NodeFilter, None] = None,
        filter: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),  # pylint: disable=redefined-builtin
        highlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
        lowlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
        enable_summary: Optional[bool] = HtmlView.PresetArgValue(None),
        max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
        enable_summary_tooltip: bool = HtmlView.PresetArgValue(True),
        enable_key_tooltip: bool = HtmlView.PresetArgValue(True),
        **kwargs,
    ) -> Html:
      """Returns the topmost HTML representation of the object.

      Args:
        view: The view to render the object.
        name: The name of the object.
        parent: The parent of the object.
        root_path: The key path of the object relative to the root.
        title: The title of the summary.
        special_keys: The special keys to display (at the immediate child
          level).
        include_keys: The keys to include (at the immediate child level).
        exclude_keys: The keys to exclude (at the immediate child level).
        collapse_level: The level to collapse the tree (relative to this node).
        uncollapse: A key path set (relative to root_path) for the nodes to
          uncollapse. or a function with signature (path, value, parent) -> bool
          to filter nodes to uncollapse.
        filter: A function with signature (path, value, parent) -> include
          to determine whether to include a field (at all levels).
        highlight: A function with signature (path, value, parent) -> bool
          to determine whether to highlight a field.
        lowlight: A function with signature (path, value, parent) -> bool
          to determine whether to lowlight a field.
        enable_summary: Whether to enable the summary. If None, summary will
          be enabled for complex types or when string exceeds
          `max_summary_len_for_str`.
        max_summary_len_for_str: The maximum length of the string to display.
        enable_summary_tooltip: Whether to enable the tooltip for the summary.
        enable_key_tooltip: Whether to enable the tooltip for the key.
        **kwargs: Additional keyword arguments passed from `pg.to_html`.

      Returns:
        The rendered HTML.
      """
      return view.render(
          self,
          name=name,
          parent=parent,
          root_path=root_path,
          title=title,
          special_keys=special_keys,
          include_keys=include_keys,
          exclude_keys=exclude_keys,
          filter=filter,
          highlight=highlight,
          lowlight=lowlight,
          enable_summary=enable_summary,
          max_summary_len_for_str=max_summary_len_for_str,
          enable_summary_tooltip=enable_summary_tooltip,
          enable_key_tooltip=enable_key_tooltip,
          collapse_level=collapse_level,
          uncollapse=uncollapse,
          **kwargs
      )

    def _html_tree_view_summary(
        self,
        *,
        view: 'HtmlTreeView',
        name: Optional[str],
        parent: Any,
        root_path: KeyPath,
        title: Union[str, Html, None] = None,
        enable_summary: Optional[bool] = HtmlView.PresetArgValue(None),
        max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
        enable_summary_tooltip: bool = HtmlView.PresetArgValue(True),
        **kwargs,
    ) -> Optional[Html]:
      """Returns the HTML representation of the object.

      Args:
        view: The view to render the object.
        name: The name of the object.
        parent: The parent of the object.
        root_path: The key path of the object relative to the root.
        title: The title of the summary.
        enable_summary: Whether to enable the summary. If None, summary will
          be enabled for complex types or when string exceeds
          `max_summary_len_for_str`.
        max_summary_len_for_str: The maximum length of the string to display.
        enable_summary_tooltip: Whether to enable the tooltip for the summary.
        **kwargs: Additional keyword arguments passed from `pg.to_html`. These
          arguments may be handled by the user logic but not the general
          HtmlTreeView.

      Returns:
        An optional HTML object representing the summary of the object. If None,
        the summary will be hidden.
      """
      return view.summary(
          self,
          name=name,
          parent=parent,
          root_path=root_path,
          title=title,
          enable_summary=enable_summary,
          max_summary_len_for_str=max_summary_len_for_str,
          enable_summary_tooltip=enable_summary_tooltip,
          **kwargs,
      )

    def _html_tree_view_content(
        self,
        *,
        view: 'HtmlTreeView',
        name: Optional[str],
        parent: Any,
        root_path: KeyPath,
        special_keys: Optional[Sequence[Union[int, str]]] = None,
        include_keys: Optional[Iterable[Union[int, str]]] = None,
        exclude_keys: Optional[Iterable[Union[int, str]]] = None,
        collapse_level: Optional[int] = 1,
        uncollapse: Union[KeyPathSet, base.NodeFilter, None] = None,
        filter: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),  # pylint: disable=redefined-builtin
        highlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
        lowlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
        max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
        enable_summary_tooltip: bool = HtmlView.PresetArgValue(True),
        enable_key_tooltip: bool = HtmlView.PresetArgValue(True),
        **kwargs,
        ) -> Html:
      """Returns the main content for the object.

      Args:
        view: The view to render the object.
        name: The name of the object.
        parent: The parent of the object.
        root_path: The key path of the object relative to the root.
        special_keys: The special keys to display (at the immediate child
          level).
        include_keys: The keys to include (at the immediate child level).
        exclude_keys: The keys to exclude (at the immediate child level).
        collapse_level: The level to collapse the tree (relative to this node).
        uncollapse: A key path set (relative to root_path) for the nodes to
          uncollapse. or a function with signature (path, value, parent) -> bool
          to filter nodes to uncollapse.
        filter: A function with signature (path, value, parent) -> include
          to determine whether to include a field (at all levels).
        highlight: A function with signature (path, value, parent) -> bool
          to determine whether to highlight a field (at all levels).
        lowlight: A function with signature (path, value, parent) -> bool
          to determine whether to lowlight a field (at all levels).
        max_summary_len_for_str: The maximum length of the string to display.
        enable_summary_tooltip: Whether to enable the tooltip for the summary.
        enable_key_tooltip: Whether to enable the key tooltip.
        **kwargs: Additional keyword arguments passed from `pg.to_html`. These
          arguments may be handled by the user logic but not the general
          HtmlTreeView.

      Returns:
        The rendered HTML as the main content of the object.
      """
      return view.content(
          self,
          name=name,
          parent=parent,
          root_path=root_path,
          special_keys=special_keys,
          include_keys=include_keys,
          exclude_keys=exclude_keys,
          filter=filter,
          highlight=highlight,
          lowlight=lowlight,
          max_summary_len_for_str=max_summary_len_for_str,
          enable_summary_tooltip=enable_summary_tooltip,
          enable_key_tooltip=enable_key_tooltip,
          collapse_level=collapse_level,
          uncollapse=uncollapse,
          **kwargs
      )

    def _html_tree_view_tooltip(
        self,
        *,
        view: 'HtmlTreeView',
        name: Optional[str],
        parent: Any,
        root_path: KeyPath,
        content: Union[str, Html, None] = None,
        **kwargs,
    ) -> Optional[Html]:
      """Returns the tooltip for the object.

      Args:
        view: The view to render the object.
        name: The referenced name of the object.
        parent: The parent of the object.
        root_path: The key path of the object relative to the root.
        content: Custom content to display in the tooltip.
        **kwargs: Additional keyword arguments passed from `pg.to_html`. These
          arguments may be handled by the user logic but not the general
          HtmlTreeView.

      Returns:
        An optional HTML object representing the tooltip of the object. If None,
        the tooltip will be hidden.
      """
      return view.tooltip(
          value=self, name=name, parent=parent, root_path=root_path,
          content=content, **kwargs
      )

  @HtmlView.extension_method('_html_tree_view_render')
  def render(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      parent: Any = None,
      root_path: Optional[KeyPath] = None,
      title: Union[str, Html, None] = None,
      special_keys: Optional[Sequence[Union[int, str]]] = None,
      include_keys: Optional[Iterable[Union[int, str]]] = None,
      exclude_keys: Optional[Iterable[Union[int, str]]] = None,
      collapse_level: Optional[int] = 1,
      uncollapse: Union[KeyPathSet, base.NodeFilter, None] = None,
      filter: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),  # pylint: disable=redefined-builtin
      highlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
      lowlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
      enable_summary: Optional[bool] = HtmlView.PresetArgValue(None),
      max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
      enable_summary_tooltip: bool = HtmlView.PresetArgValue(True),
      enable_key_tooltip: bool = HtmlView.PresetArgValue(True),
      **kwargs
  ) -> Html:
    """Renders the entire HTML tree view for the value.

    Args:
      value: The value to render.
      name: The name of the value.
      parent: The parent of the value.
      root_path: The root path of the value.
      title: The title of the summary.
      special_keys: The special keys to display (at the immediate child level).
      include_keys: The keys to include (at the immediate child level).
      exclude_keys: The keys to exclude (at the immediate child level).
      collapse_level: The level to collapse the tree (relative to this node).
      uncollapse: A key path set (relative to root_path) for the nodes to
        uncollapse. or a function with signature (path, value, parent) -> bool
        to filter nodes to uncollapse.
      filter: A function with signature (path, value, parent) -> include
        to determine whether to include a field (at all levels).
      highlight: A function with signature (path, value, parent) -> bool
        to determine whether to highlight a field.
      lowlight: A function with signature (path, value, parent) -> bool
        to determine whether to lowlight a field.
      enable_summary: Whether to enable the summary. If None, summary will
        be enabled for complex types or when string exceeds
        `max_summary_len_for_str`.
      max_summary_len_for_str: The maximum length of the string to display.
      enable_summary_tooltip: Whether to enable the tooltip for the summary.
      enable_key_tooltip: Whether to enable the key tooltip.
      **kwargs: Additional keyword arguments passed from `pg.to_html`.

    Returns:
      The rendered HTML.
    """
    root_path = root_path or KeyPath()
    uncollapse = self.init_uncollapse(uncollapse)

    if isinstance(value, HtmlTreeView.Extension):
      extension_collapse_level = value._html_tree_view_collapse_level()    # pylint: disable=protected-access

      # If the extension has child levels to uncollapse, honor them above the
      # collapse level passed from the root. However, we can see the
      # uncollapsed extension subtree only when the extension's parent node is
      # uncollapsed.
      if (extension_collapse_level > 0
          and root_path.depth + extension_collapse_level > collapse_level):
        collapse_level = root_path.depth + extension_collapse_level

      if not callable(uncollapse):
        extension_uncollapse = value._html_tree_view_uncollapse().copy()  # pylint: disable=protected-access
        if extension_uncollapse:
          extension_uncollapse.rebase(root_path)
          uncollapse = uncollapse.union(extension_uncollapse)

    summary = self.summary(
        value,
        name=name,
        parent=parent,
        root_path=root_path,
        title=title,
        enable_summary=enable_summary,
        enable_summary_tooltip=enable_summary_tooltip,
        max_summary_len_for_str=max_summary_len_for_str,
        **kwargs,
    )
    content = self.content(
        value,
        name=name,
        parent=parent,
        root_path=root_path,
        filter=filter,
        special_keys=special_keys,
        include_keys=include_keys,
        exclude_keys=exclude_keys,
        collapse_level=collapse_level,
        uncollapse=uncollapse,
        max_summary_len_for_str=max_summary_len_for_str,
        enable_summary_tooltip=enable_summary_tooltip,
        enable_key_tooltip=enable_key_tooltip,
        **kwargs,
    )
    extension_style = (
        value._html_style() if isinstance(value, HtmlView.Extension) else []  # pylint: disable=protected-access
    )

    if summary is None:
      content = Html.from_value(content)
      assert content is not None
      content.add_style(*extension_style)
      return content

    collapse_view = self.should_collapse(
        value, parent=parent, root_path=root_path,
        collapse_level=collapse_level, uncollapse=uncollapse,
    )
    return Html.element(
        'details',
        [
            summary,
            content,
        ],
        options=[None if collapse_view else 'open'],
        css_class=[
            'pyglove',
            self.css_class_name(value),
        ],
    ).add_style(
        """
        /* Value details styles. */
        details.pyglove {
          border: 1px solid #aaa;
          border-radius: 4px;
          padding: 0.5em 0.5em 0;
          margin: 0.1em 0;
        }
        details.pyglove.special_value {
          margin-bottom: 0.75em;
        }
        details.pyglove[open] {
          padding: 0.5em 0.5em 0.5em;
        }
        .highlight {
          background-color: Mark;
        }
        .lowlight {
          opacity: 0.2;
        }
        """,
        *extension_style,
    )

  def init_uncollapse(
      self,
      uncollapse: Union[
          Iterable[Union[KeyPath, str]], base.NodeFilter, None
      ] = HtmlView.PresetArgValue(None),
  ) -> Union[KeyPathSet, base.NodeFilter]:
    """Normalize the uncollapse argument."""
    if uncollapse is None:
      return KeyPathSet()
    elif callable(uncollapse):
      return uncollapse
    else:
      return KeyPathSet.from_value(uncollapse, include_intermediate=True)

  def should_collapse(
      self,
      value: Any,
      root_path: KeyPath,
      parent: Any,
      collapse_level: Optional[int] = 0,
      uncollapse: Union[KeyPathSet, base.NodeFilter] = None,
  ) -> bool:
    """Returns whether the object should be collapsed."""
    if collapse_level is None or root_path.depth < collapse_level:
      return False
    if callable(uncollapse):
      return not uncollapse(root_path, value, parent)
    else:
      return root_path not in uncollapse

  def needs_summary(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      parent: Any = None,
      title: Union[str, Html, None] = None,
      enable_summary: Optional[bool] = HtmlView.PresetArgValue(True),
      max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
  ) -> bool:
    """Returns whether the object needs a summary."""
    del parent
    if enable_summary is None:
      if name is not None or title is not None or not (
          isinstance(value, (int, float, bool, type(None)))
          or (
              isinstance(value, str)
              and len(value) <= max_summary_len_for_str
          )
      ):
        return True
    return enable_summary

  @HtmlView.extension_method('_html_tree_view_summary')
  def summary(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      parent: Any = None,
      root_path: Optional[KeyPath] = None,
      title: Union[str, Html, None] = None,
      enable_summary: Optional[bool] = HtmlView.PresetArgValue(None),
      enable_summary_tooltip: bool = HtmlView.PresetArgValue(True),
      max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
      **kwargs
  ) -> Optional[Html]:
    """Renders a summary for the value.

    Args:
      value: The value to render.
      name: The name of the value.
      parent: The parent of the value.
      root_path: The root path of the value.
      title: The title of the summary.
      enable_summary: Whether to enable the summary. If None, summary will
        be enabled for complex types or when string exceeds
        `max_summary_len_for_str`.
      enable_summary_tooltip: Whether to enable the tooltip for the summary.
      max_summary_len_for_str: The maximum length of the string to display.
      **kwargs: Additional keyword arguments passed from `pg.to_html`.

    Returns:
      An optional HTML object representing the summary of the value. If None,
      the summary will be hidden.
    """
    if not self.needs_summary(
        value,
        name=name,
        parent=parent,
        title=title,
        max_summary_len_for_str=max_summary_len_for_str,
        enable_summary=enable_summary,
    ):
      return None

    def make_title(value: Any):
      if isinstance(value, str):
        if len(value) > max_summary_len_for_str:
          value = value[:max_summary_len_for_str] + '...'
        return Html.escape(repr(value))
      return f'{type(value).__name__}(...)'

    return Html.element(
        'summary',
        [
            # Summary name.
            lambda: Html.element(  # pylint: disable=g-long-ternary
                'div',
                [
                    name,
                ],
                css_class=['summary_name']
            ) if name else None,

            # Summary title
            Html.element(
                'div',
                [
                    title or make_title(value),
                ],
                css_class=['summary_title']
            ),

            # Tooltip.
            lambda: self.tooltip(   # pylint: disable=g-long-ternary
                value,
                name=name,
                parent=parent,
                root_path=root_path,
                **kwargs,
            ) if enable_summary_tooltip else None,
        ],
    ).add_style(
        """
        /* Summary styles. */
        details.pyglove summary {
          font-weight: bold;
          margin: -0.5em -0.5em 0;
          padding: 0.5em;
        }
        .summary_name {
          display: inline;
          padding: 0 5px;
        }
        .summary_title {
          display: inline;
        }
        .summary_name + div.summary_title {
          display: inline;
          color: #aaa;
        }
        .summary_title:hover + span.tooltip {
          visibility: visible;
        }
        /* Type-specific styles. */
        .pyglove.str .summary_title {
          color: darkred;
          font-style: italic;
        }
        """
    )

  # NOTE(daiyip)" `object_key`` does not have a corresponding extension
  # method in `HtmlTreeView.Extension`, because the rendering of the key is not
  # delegated to `HtmlTreeView.Extension`.
  def object_key(
      self,
      key: Union[str, int],
      *,
      name: Optional[str] = None,
      parent: Any,
      root_path: Optional[KeyPath] = None,
      css_class: Union[str, Sequence[str], None] = None,
      key_color: Optional[str] = None,
      enable_tooltip: bool = HtmlView.PresetArgValue(True),
      **kwargs
  ) -> Html:
    """Renders the key for the value.

    Args:
      key: The key of the value.
      name: The name of the value.
      parent: The parent value of the key.
      root_path: The root path of the value.
      css_class: Additional CSS classes to add to the HTML element.
      key_color: The color of the key.
      enable_tooltip: Whether to enable the tooltip.
      **kwargs: Additional keyword arguments passed from `pg.to_html`.

    Returns:
      The rendered HTML as the key of the value.
    """
    return (
        # Key span.
        Html.element(
            'span',
            [
                str(key),
            ],
            css_class=[
                'object_key',
                type(key).__name__,
                css_class,
            ],
            style=dict(
                color=key_color,
            )
        ) + (
            # Tooltip if enabled.
            lambda: self.tooltip(  # pylint: disable=g-long-ternary
                value=root_path,
                root_path=root_path,
                name=name,
                parent=parent,
                **kwargs
            ) if enable_tooltip else None
        )
    ).add_style(
        """
        /* Object key styles. */
        .object_key {
          margin-right: 0.25em;
        }
        .object_key:hover + .tooltip {
          visibility: visible;
          background-color: darkblue;
        }
        .object_key.str {
          color: gray;
          border: 1px solid lightgray;
          background-color: ButtonFace;
          border-radius: 0.2em;
          padding: 0.3em;
        }
        .object_key.int::before{
          content: '[';
        }
        .object_key.int::after{
          content: ']';
        }
        .object_key.int{
          border: 0;
          color: lightgray;
          background-color: transparent;
          border-radius: 0;
          padding: 0;
        }
        """
    )

  @HtmlView.extension_method('_html_tree_view_content')
  def content(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      parent: Any = None,
      root_path: Optional[KeyPath] = None,
      special_keys: Optional[Sequence[Union[int, str]]] = None,
      include_keys: Optional[Iterable[Union[int, str]]] = None,
      exclude_keys: Optional[Iterable[Union[int, str]]] = None,
      collapse_level: Optional[int] = 1,
      uncollapse: Union[KeyPathSet, base.NodeFilter, None] = None,
      filter: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),  # pylint: disable=redefined-builtin
      highlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
      lowlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
      max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
      enable_summary_tooltip: bool = HtmlView.PresetArgValue(True),
      enable_key_tooltip: bool = HtmlView.PresetArgValue(True),
      **kwargs
  ) -> Html:
    """Renders the main content for the value.

    Args:
      value: The value to render.
      name: The name of the value.
      parent: The parent of the value.
      root_path: The root path of the value.
      special_keys: The special keys to display (at the immediate child level).
      include_keys: The keys to include (at the immediate child level).
      exclude_keys: The keys to exclude (at the immediate child level).
      collapse_level: The level to collapse the tree (relative to this node).
      uncollapse: A key path set (relative to root_path) for the nodes to
        uncollapse. or a function with signature (path, value, parent) -> bool
        to filter nodes to uncollapse.
      filter: A function with signature (path, value, parent) -> include
        to determine whether to include a field (at all levels).
      highlight: A function with signature (path, value, parent) -> bool
        to determine whether to highlight a field (at all levels).
      lowlight: A function with signature (path, value, parent) -> bool
        to determine whether to lowlight a field (at all levels).
      max_summary_len_for_str: The maximum length of the string to display.
      enable_summary_tooltip: Whether to enable the summary tooltip.
      enable_key_tooltip: Whether to enable the key tooltip.
      **kwargs: Additional keyword arguments passed from `pg.to_html`.

    Returns:
      The rendered HTML as the main content of the value.
    """
    if isinstance(value, (tuple, list)):
      items = {i: v for i, v in enumerate(value)}
    elif isinstance(value, dict):
      items = value
    else:
      return self.simple_value(
          value, name=name, parent=parent, root_path=root_path,
          max_summary_len_for_str=max_summary_len_for_str
      )
    return self.complex_value(
        items,
        name=name,
        parent=value,
        root_path=root_path or KeyPath(),
        special_keys=special_keys,
        include_keys=include_keys,
        exclude_keys=exclude_keys,
        collapse_level=collapse_level,
        uncollapse=uncollapse,
        filter=filter,
        highlight=highlight,
        lowlight=lowlight,
        max_summary_len_for_str=max_summary_len_for_str,
        enable_summary_tooltip=enable_summary_tooltip,
        enable_key_tooltip=enable_key_tooltip,
        **kwargs,
    )

  def simple_value(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      parent: Any = None,
      root_path: Optional[KeyPath] = None,
      css_class: Union[str, Sequence[str], None] = None,
      max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
  ) -> Html:
    """Renders a simple value.

    Args:
      value: The value to render.
      name: The name of the value.
      parent: The parent of the value.
      root_path: The root path of the value.
      css_class: Additional CSS classes to add to the HTML element.
      max_summary_len_for_str: The maximum length of the string to display.

    Returns:
      The rendered HTML as the simple value.
    """
    del name, parent, root_path
    def value_repr() -> str:
      if isinstance(value, str):
        if len(value) < max_summary_len_for_str:
          return repr(value)
        else:
          return value
      return object_utils.format(
          value,
          compact=False, verbose=False, hide_default_values=True,
          python_format=True, use_inferred=True,
          max_bytes_len=64,
      )
    return Html.element(
        'span',
        [
            Html.escape(value_repr),
        ],
        css_class=[
            'simple_value',
            self.css_class_name(value),
            css_class,
        ],
    ).add_style(
        """
        /* Simple value styles. */
        .simple_value {
          color: blue;
          display: inline-block;
          white-space: pre-wrap;
          padding: 0.2em;
          margin-top: 0.15em;
        }
        .simple_value.str {
          color: darkred;
          font-style: italic;
        }
        .simple_value.int, .simple_value.float {
          color: darkblue;
        }
        """
    )

  def complex_value(
      self,
      kv: Dict[Union[int, str], Any],
      *,
      parent: Any,
      root_path: KeyPath,
      name: Optional[str] = None,
      css_class: Union[str, Sequence[str], None] = None,
      render_key_fn: Optional[Callable[..., Html]] = None,
      render_value_fn: Optional[Callable[..., Html]] = None,
      special_keys: Optional[Sequence[Union[int, str]]] = None,
      include_keys: Optional[Iterable[Union[int, str]]] = None,
      exclude_keys: Optional[Iterable[Union[int, str]]] = None,
      collapse_level: Optional[int] = 1,
      uncollapse: Union[KeyPathSet, base.NodeFilter, None] = None,
      filter: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),  # pylint: disable=redefined-builtin
      highlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
      lowlight: Optional[base.NodeFilter] = HtmlView.PresetArgValue(None),
      max_summary_len_for_str: int = HtmlView.PresetArgValue(40),
      enable_summary_tooltip: bool = HtmlView.PresetArgValue(True),
      enable_key_tooltip: bool = HtmlView.PresetArgValue(True),
      **kwargs,
  ) -> Html:
    """Renders a list of key-value pairs.

    Args:
      kv: The key-value pairs to render.
      parent: The parent value of the key-value pairs.
      root_path: The root path of the value.
      name: The name of the value.
      css_class: Additional CSS classes to add to the HTML element.
      render_key_fn: A function to render the key. The function has the
        same signature as `HtmlTreeView.object_key`.
        If None, `HtmlTreeView.object_key` will be used to render the key.
      render_value_fn: A function to render the value. The function has the
        same signature as `HtmlTreeView.render`.
        If None, `HtmlTreeView.render` will be used to render child value.
      special_keys: The special keys to display (at the immediate child level).
      include_keys: The keys to include (at the immediate child level).
      exclude_keys: The keys to exclude (at the immediate child level).
      collapse_level: The level to collapse the tree (relative to this node).
      uncollapse: A key path set (relative to root_path) for the nodes to
        uncollapse. or a function with signature (path, value, parent) -> bool
        to filter nodes to uncollapse.
      filter: A function with signature (path, value, parent) -> include
        to determine whether to include.
      highlight: A function with signature (path, value, parent) -> bool
        to determine whether to highlight.
      lowlight: A function with signature (path, value, parent) -> bool
        to determine whether to lowlight.
      max_summary_len_for_str: The maximum length of the string to display.
      enable_summary_tooltip: Whether to enable the summary tooltip.
      enable_key_tooltip: Whether to enable the key tooltip.
      **kwargs: Additional keyword arguments passed from `pg.to_html`.

    Returns:
      The rendered HTML as the key-value pairs.
    """
    del name
    root_path = root_path or KeyPath()
    uncollapse = self.init_uncollapse(uncollapse)

    if isinstance(parent, HtmlTreeView.Extension):
      special_keys = special_keys or parent._html_tree_view_special_keys()  # pylint: disable=protected-access
      include_keys = include_keys or parent._html_tree_view_include_keys()  # pylint: disable=protected-access
      exclude_keys = exclude_keys or parent._html_tree_view_exclude_keys()  # pylint: disable=protected-access

    special_keys = special_keys or []
    include_keys = set(include_keys or [])
    exclude_keys = set(exclude_keys or [])

    render_key_fn = render_key_fn or self.object_key
    render_value_fn = render_value_fn or self.render

    s = Html()
    if kv:
      include_keys = include_keys or set(kv.keys())
      if filter is not None:
        include_keys -= set(
            k for k, v in kv.items()
            if not filter(root_path + k, v, parent)
        )
      if exclude_keys:
        include_keys -= exclude_keys

      if special_keys:
        for k in special_keys:
          if k in include_keys and k in kv:
            child_path = root_path + k
            v = kv[k]
            s.write(
                Html.element(
                    'div',
                    [
                        render_value_fn(
                            value=v,
                            name=k,
                            parent=parent,
                            root_path=child_path,
                            filter=filter,
                            special_keys=None,
                            include_keys=None,
                            exclude_keys=None,
                            collapse_level=collapse_level,
                            uncollapse=uncollapse,
                            highlight=highlight,
                            lowlight=lowlight,
                            max_summary_len_for_str=max_summary_len_for_str,
                            enable_summary_tooltip=enable_summary_tooltip,
                            enable_key_tooltip=enable_key_tooltip,
                            **kwargs
                        )
                    ],
                    css_class=[
                        'special_value',
                        (
                            'highlight' if highlight
                            and highlight(child_path, v, parent) else None
                        ),
                        (
                            'lowlight' if lowlight
                            and lowlight(child_path, v, parent) else None
                        )
                    ],
                )
            )
            include_keys.remove(k)

      if include_keys:
        s.write('<table>')
        for k, v in kv.items():
          if k not in include_keys:
            continue
          child_path = root_path + k
          key_cell = render_key_fn(
              key=k,
              parent=parent,
              root_path=child_path,
              enable_tooltip=enable_key_tooltip,
          )
          value_cell = Html.element(
              'div',
              [
                  render_value_fn(
                      value=v,
                      name=None,
                      parent=parent,
                      root_path=child_path,
                      special_keys=None,
                      include_keys=None,
                      exclude_keys=None,
                      collapse_level=collapse_level,
                      uncollapse=uncollapse,
                      filter=filter,
                      highlight=highlight,
                      lowlight=lowlight,
                      max_summary_len_for_str=max_summary_len_for_str,
                      enable_summary_tooltip=enable_summary_tooltip,
                      enable_key_tooltip=enable_key_tooltip,
                      **kwargs,
                  )
              ],
              css_class=[
                  (
                      'highlight' if highlight
                      and highlight(child_path, v, parent) else None
                  ),
                  (
                      'lowlight' if lowlight
                      and lowlight(child_path, v, parent) else None
                  )
              ],
          )
          s.write(
              Html.element(
                  'tr',
                  [
                      '<td>', key_cell, '</td>',
                      '<td>', value_cell, '</td>',
                  ],
              )
          )
        s.write('</table>')
    else:
      s.write(Html.element('span', css_class=['empty_container']))
    return Html.element(
        'div',
        [s],
        css_class=[
            'complex_value',
            self.css_class_name(parent),
            css_class,
        ]
    ).add_style(
        """
        /* Complex value styles. */
        span.empty_container::before {
            content: '(empty)';
            font-style: italic;
            margin-left: 0.5em;
            color: #aaa;
        }
        """
    )

  @HtmlView.extension_method('_html_tree_view_tooltip')
  def tooltip(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      parent: Any = None,
      root_path: Optional[KeyPath] = None,
      content: Union[str, Html, None] = HtmlView.PresetArgValue(None),
      **kwargs
  ) -> Html:
    """Renders a tooltip for the value.

    Args:
      value: The value to render.
      name: The name of the value.
      parent: The parent value of the key-value pairs.
      root_path: The root path of the value.
      content: The content of the tooltip. If None, the formatted value will be
        used as the content.
      **kwargs: Additional keyword arguments passed from `pg.to_html`.

    Returns:
      The rendered HTML as the tooltip of the value.
    """
    del name, parent
    if content is None:
      content = Html.escape(
          object_utils.format(
              value,
              root_path=root_path,
              compact=False,
              verbose=False,
              python_format=True,
              max_bytes_len=64,
              max_str_len=256,
              **kwargs
          )
      )
    return Html.element(
        'span',
        [content],
        css_class=['tooltip', self.css_class_name(value)],
    ).add_style(
        """
        /* Tooltip styles. */
        span.tooltip {
          visibility: hidden;
          white-space: pre-wrap;
          font-weight: normal;
          background-color: #484848;
          color: #fff;
          padding: 10px;
          border-radius: 6px;
          position: absolute;
          z-index: 1;
        }
        """
    )

  @staticmethod
  def css_class_name(value: Any) -> Optional[str]:
    """Returns the CSS class name for the value."""
    if isinstance(value, HtmlTreeView.Extension):
      return Html.concate(value._html_element_class())  # pylint: disable=protected-access
    value = value if inspect.isclass(value) else type(value)
    return object_utils.camel_to_snake(value.__name__, '-')


_REGEX_CAMEL_TO_SNAKE = re.compile(r'(?<!^)(?=[A-Z])')

# pytype: enable=annotation-type-mismatch



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools
import os
import tempfile
from typing import Any, Callable, Iterable, Optional, Union
import unittest

from pyglove.core import io as pg_io
from pyglove.core.views import base

Content = base.Content
View = base.View


class ReferenceLinks(Content.SharedParts):

  @functools.cached_property
  def content(self) -> str:
    return '\n'.join(
        [
            f'link={url}' for url in self.parts.keys()
        ]
    )


class SharedPartsTest(unittest.TestCase):

  def test_basics(self):
    links = ReferenceLinks()
    self.assertFalse(links)
    self.assertEqual(links.content, '')
    self.assertEqual(list(links), [])

    links = ReferenceLinks('https://x/y.css')
    self.assertEqual(links.content, 'link=https://x/y.css')
    self.assertTrue(links)
    self.assertIn('https://x/y.css', links)
    self.assertNotIn('https://x/z.css', links)
    self.assertEqual(list(links), ['https://x/y.css'])

    links = ReferenceLinks('https://x/y.css', 'https://x/z.css')
    self.assertEqual(
        links.content,
        'link=https://x/y.css\nlink=https://x/z.css'
    )
    self.assertTrue(links)
    self.assertIn('https://x/y.css', links)
    self.assertIn('https://x/z.css', links)
    self.assertEqual(list(links), ['https://x/y.css', 'https://x/z.css'])

  def test_add(self):
    links = ReferenceLinks('https://x/y.css')
    self.assertEqual(links.content, 'link=https://x/y.css')

    # Updated.
    self.assertTrue(links.add('https://x/z.css'))
    self.assertEqual(links.parts, {'https://x/y.css': 1, 'https://x/z.css': 1})
    # Make sure the cached property `content` is updated.
    self.assertEqual(
        links.content,
        'link=https://x/y.css\nlink=https://x/z.css'
    )
    # Not updated but the number of reference gets updated.
    self.assertFalse(links.add('https://x/z.css'))
    self.assertEqual(links.parts, {'https://x/y.css': 1, 'https://x/z.css': 2})

    # Adding None is ignored.
    self.assertFalse(links.add(None))
    self.assertEqual(links.parts, {'https://x/y.css': 1, 'https://x/z.css': 2})

    # Add another instance of `ReferenceLinks`.
    self.assertTrue(
        links.add(ReferenceLinks('https://x/y.css', 'https://x/w.css'))
    )
    self.assertEqual(
        links.parts,
        {
            'https://x/y.css': 2,
            'https://x/z.css': 2,
            'https://x/w.css': 1
        }
    )

  def test_copy(self):
    links = ReferenceLinks('https://x/y.css')
    links2 = copy.copy(links)
    self.assertEqual(links, links2)
    self.assertEqual(links.parts, links2.parts)

  def test_eq(self):
    links1 = ReferenceLinks('https://x/y.css')
    links2 = ReferenceLinks('https://x/y.css')
    self.assertEqual(links1, links2)
    self.assertNotEqual(links1, 'abc')
    self.assertNotEqual(links1, ReferenceLinks('https://x/z.css'))

  def test_format(self):
    links = ReferenceLinks('https://x/y.css')
    self.assertEqual(
        repr(links),
        """ReferenceLinks(parts={'https://x/y.css': 1})"""
    )
    self.assertEqual(
        str(links),
        """link=https://x/y.css"""
    )


class Document(Content):

  WritableTypes = Union[    # pylint: disable=invalid-name
      str, 'Document', None, Callable[[], Union[str, 'Document', None]]
  ]

  def __init__(   # pylint: disable=useless-super-delegation
      self,
      *content: WritableTypes, ref_links: Optional[Iterable[str]] = None):
    super().__init__(*content, ref_links=ReferenceLinks(*(ref_links or [])))

  @property
  def ref_links(self) -> ReferenceLinks:
    return self._shared_parts['ref_links']

  def to_str(self, *, content_only: bool = False, **kwargs):
    if content_only:
      return self.content
    return self.ref_links.content + '\n' + self.content


class ContentTest(unittest.TestCase):

  def test_basics(self):
    document = Document()
    self.assertEqual(document.content, '')

    document = Document('abc', 'def', ref_links=['https://x/y.css'])
    self.assertEqual(document.content, 'abcdef')
    self.assertEqual(document.to_str(content_only=True), 'abcdef')
    self.assertEqual(document.to_str(), 'link=https://x/y.css\nabcdef')
    self.assertEqual(document.shared_parts, {'ref_links': document.ref_links})
    self.assertEqual(
        Document('abc', ref_links=['https://x/z.css']),
        Document('abc', ref_links=['https://x/z.css'])
    )
    self.assertNotEqual(
        Document('abc', ref_links=['https://x/y.css']),
        'abc'
    )
    self.assertNotEqual(
        Document('abc', ref_links=['https://x/y.css']),
        Document('abc', ref_links=['https://x/z.css'])
    )
    self.assertEqual(
        hash(Document('abc', ref_links=['https://x/y.css'])),
        hash(Document('abc', ref_links=['https://x/y.css']))
    )
    self.assertNotEqual(
        hash(Document('abc', ref_links=['https://x/y.css'])),
        hash(Document('abc', ref_links=['https://x/z.css']))
    )

  def test_write(self):
    document = Document('abc', ref_links=['https://x/y.css'])
    self.assertEqual(document.content, 'abc')

    self.assertIs(document.write('def'), document)
    self.assertEqual(document.content, 'abcdef')

    # No op.
    document.write(None)
    self.assertEqual(document.content, 'abcdef')

    # Write another instance of `Document`.
    document.write(
        Document('ghi', ref_links=['https://x/y.css', 'https://x/z.css'])
    )
    self.assertEqual(document.content, 'abcdefghi')
    self.assertEqual(
        document.to_str(),
        'link=https://x/y.css\nlink=https://x/z.css\nabcdefghi'
    )

    # Write a lambda.
    document.write(lambda: 'jkl')
    self.assertEqual(document.content, 'abcdefghijkl')
    self.assertEqual(
        document.to_str(),
        'link=https://x/y.css\nlink=https://x/z.css\nabcdefghijkl'
    )

    document.write(lambda: None)
    self.assertEqual(document.content, 'abcdefghijkl')

    document.write(lambda: Document('mno'))
    self.assertEqual(document.content, 'abcdefghijklmno')

  def test_save(self):
    filename = os.path.join(tempfile.gettempdir(), '1', 'test_doc.txt')
    Document('abc', ref_links=['https://x/y.css']).save(filename)
    self.assertTrue(pg_io.path_exists(filename))
    self.assertEqual(pg_io.readfile(filename), 'link=https://x/y.css\nabc')

  def test_add_and_radd(self):
    self.assertEqual(
        (Document('abc', ref_links=['https://x/y.css']) + 'def').to_str(),
        'link=https://x/y.css\nabcdef'
    )
    self.assertEqual(
        ('def' + Document('abc', ref_links=['https://x/y.css'])).to_str(),
        'link=https://x/y.css\ndefabc'
    )
    self.assertEqual(
        (Document('abc', ref_links=['https://x/y.css']) + None).to_str(),
        'link=https://x/y.css\nabc'
    )
    self.assertEqual(
        (None + Document('abc', ref_links=['https://x/y.css'])).to_str(),
        'link=https://x/y.css\nabc'
    )
    self.assertEqual(
        (Document('abc', ref_links=['https://x/y.css'])
         + (lambda: 'def')).to_str(),
        'link=https://x/y.css\nabcdef'
    )
    self.assertEqual(
        ((lambda: 'def')
         + Document('abc', ref_links=['https://x/y.css'])).to_str(),
        'link=https://x/y.css\ndefabc'
    )

  def test_format(self):
    document = Document('abc', ref_links=['https://x/y.css'])
    self.assertEqual(
        repr(document),
        (
            'Document(content=\'abc\', '
            'ref_links=ReferenceLinks(parts={\'https://x/y.css\': 1}))'
        )
    )
    self.assertEqual(
        str(document),
        'link=https://x/y.css\nabc'
    )

  def test_from_value(self):
    document = Document('abc', ref_links=['https://x/y.css'])
    self.assertIs(Document.from_value(document), document)
    document2 = Document.from_value(document, copy=True)
    self.assertIsNot(document2, document)
    self.assertEqual(document2, document)

    self.assertIsNone(Document.from_value(None))
    self.assertEqual(Document.from_value('abc'), Document('abc'))
    self.assertEqual(
        Document.from_value(lambda: 'abc'), Document('abc')
    )


class TestView(View):

  VIEW_ID = 'test'

  class Extension(View.Extension):

    def __init__(self, x: int):
      super().__init__()
      self.x = x

    def __str__(self):
      return f'{self.__class__.__name__}(x={self.x})'

    def _test_view_render(
        self,
        *,
        view: 'TestView',
        use_summary: bool = View.PresetArgValue(False),
        use_content: bool = View.PresetArgValue(False),
        **kwargs):
      if use_summary:
        return view.summary(self, **kwargs)
      if use_content:
        return view.content(self, **kwargs)
      return Document(f'[Custom Document] {self}')

    def _test_view_summary(
        self,
        view: 'TestView',
        default_behavior: bool = View.PresetArgValue(False),
        **kwargs):
      if default_behavior:
        return view.summary(self, **kwargs)
      return Document(f'[Custom Summary] {self}')

    def _test_view_content(self, **kwargs):
      del kwargs
      return Document(f'[Custom Content] {self}')

    def bar(self, x: int = View.PresetArgValue(2)):
      return x

  @View.extension_method('_test_view_render')
  def render(
      self,
      value: Any,
      *,
      greeting: str = View.PresetArgValue('hi'),
      **kwargs):
    return Document(f'{greeting} {value}')

  @View.extension_method('_test_view_summary')
  def summary(
      self,
      value: Any,
      title: str = View.PresetArgValue('abc'),
      **kwargs
  ):
    del kwargs
    return Document(f'[Default Summary] {title}: {value}')

  @View.extension_method('_test_view_content')
  def content(self, value: Any, **kwargs):
    del kwargs
    return Document(f'[Default Content] {value}')

  def render_key(self, key: str, *, value: Optional[Any] = None, **kwargs):
    del kwargs
    return Document(f'[Default Key] {key}: {value}')

  def foo(self, x: int = View.PresetArgValue(1)):
    return x


class TestView2(View):

  VIEW_ID = 'test2'

  class Extension(View.Extension):

    def _test_view2_render(self, value: Any, **kwargs):
      del kwargs
      return Document(f'<h2>{value}</h2>')

    def _test_view2_tooltip(
        self,
        **kwargs,
    ):
      del kwargs
      return Document('<h2>Tooltip</h2>')

  @View.extension_method('_test_view2_render')
  def render(self, value: Any, **kwargs):
    return Document(f'<h1>{value}</h1>')

  @View.extension_method('_test_view2_tooltip')
  def tooltip(self, *, value: Optional[Any] = None, **kwargs):
    del kwargs
    return Document(f'<tooltip>{value}</tooltip>')


class MyObject(TestView.Extension, TestView2.Extension):

  def __init__(self, x: int, y: str):
    super().__init__(x)
    self.y = y

  def _test_view_content(
      self, *, text: str = View.PresetArgValue('MyObject'), **kwargs
  ):
    return Document(f'[{text}] {self.x} {self.y}')

  def _test_view2_render(self, **kwargs):
    return Document(f'<span>{self.x}</span>')


class MyObject2(MyObject):

  def _doc_content(self, **kwargs):
    del kwargs
    return Document('Nothing')


class ViewTest(unittest.TestCase):

  def test_bad_views(self):
    # VIEW ID must be present for concrete View classes.
    with self.assertRaisesRegex(
        ValueError, '`VIEW_ID` must be set'
    ):
      class NoTypeNonAbstractView(View):  # pylint: disable=unused-variable

        def render(self, *args, **kwargs):
          return Document('foo')

    with self.assertRaisesRegex(
        TypeError, 'must have a `value` argument'
    ):
      class NodeMethodWithoutValue(View):  # pylint: disable=unused-variable

        def render(self, *args, **kwargs):
          return Document('foo')

        @View.extension_method('_some_method')
        def bad_method(self, text: str):
          pass

    with self.assertRaisesRegex(
        TypeError, 'must not have variable positional argument.'
    ):
      class NodeMethodWithVarargs(View):  # pylint: disable=unused-variable

        def render(self, *args, **kwargs):
          return Document('foo')

        @View.extension_method('_some_method')
        def bad_method(self, value, *args):
          pass

    class AbstractView(View):
      VIEW_ID = 'abstract'

    with self.assertRaisesRegex(
        ValueError, 'The `VIEW_ID` .* is the same as the base class'
    ):
      class AbstractViewWithNodeMethod(AbstractView):  # pylint: disable=unused-variable

        def render(self, *args, **kwargs):
          return Document('foo')

  def test_create(self):
    view = View.create('test')
    self.assertIsInstance(view, TestView)
    with self.assertRaisesRegex(ValueError, 'No view class found'):
      View.create('nonexisted_view')

  def test_dir(self):
    self.assertIn('test', View.dir())

  def test_supported_view_classes(self):
    self.assertEqual(
        TestView.Extension.supported_view_classes(),
        {TestView}
    )
    self.assertEqual(
        MyObject.supported_view_classes(),
        {TestView, TestView2}
    )

  def test_supported_preset_args(self):
    self.assertEqual(
        TestView.supported_preset_args(),
        {
            TestView: {
                'greeting': View.PresetArgValue('hi'),
                'title': View.PresetArgValue('abc'),
                'x': View.PresetArgValue(1),
            },
            TestView.Extension: {
                'default_behavior': View.PresetArgValue(default=False),
                'use_content': View.PresetArgValue(False),
                'use_summary': View.PresetArgValue(False),
                'x': View.PresetArgValue(2),
            },
            MyObject: {
                'text': View.PresetArgValue('MyObject'),
            },
        }
    )
    self.assertEqual(
        TestView2.supported_preset_args(),
        {
            TestView2: {},
            MyObject: {
                'text': View.PresetArgValue('MyObject'),
            },
        }
    )

    with self.assertRaisesRegex(ValueError, 'No `VIEW_ID` is set'):
      View.supported_preset_args()

  def test_view_method_using_preset_args(self):
    view = View.create('test')
    self.assertEqual(view.foo(), 1)
    with View.preset_args(x=2):
      self.assertEqual(view.foo(), 2)

  def test_node_method_using_preset_args(self):
    self.assertEqual(MyObject(1, '').bar(), 2)
    with View.preset_args(x=3):
      self.assertEqual(MyObject(1, '').bar(), 3)

  def test_view_default_behavior(self):
    view = View.create('test')
    self.assertEqual(
        view.render(1).content,
        'hi 1'
    )
    self.assertEqual(
        view.summary(value=1).content,
        '[Default Summary] abc: 1'
    )
    self.assertEqual(
        view.content(1).content,
        '[Default Content] 1'
    )
    self.assertEqual(
        view.render_key(1).content,
        '[Default Key] 1: None'
    )
    with self.assertRaisesRegex(
        ValueError, 'No value is provided for the `value` argument'
    ):
      view.render()

  def test_custom_behavior(self):
    view = View.create('test')
    self.assertEqual(
        view.render(MyObject(1, 'foo'), use_content=True).content,
        '[MyObject] 1 foo'
    )

    self.assertEqual(
        view.render(TestView.Extension(1)).content,
        '[Custom Document] Extension(x=1)'
    )
    self.assertEqual(
        view.summary(TestView.Extension(1)).content,
        '[Custom Summary] Extension(x=1)'
    )
    self.assertEqual(
        view.content(TestView.Extension(1)).content,
        '[Custom Content] Extension(x=1)'
    )
    self.assertEqual(
        view.render_key('foo', value=TestView.Extension(1)).content,
        '[Default Key] foo: Extension(x=1)'
    )

    view2 = View.create('test2')
    self.assertEqual(
        view2.render(TestView.Extension(1)).content,
        '<h1>Extension(x=1)</h1>'
    )
    self.assertEqual(
        view2.render(MyObject(1, 'foo')).content,
        '<span>1</span>'
    )
    self.assertEqual(
        view2.tooltip(value=MyObject(1, 'foo')).content,
        '<h2>Tooltip</h2>'
    )

  def test_advanced_routing_with_arg_passing(self):
    view = View.create('test')
    self.assertEqual(
        view.render(
            TestView.Extension(1), use_summary=True).content,
        '[Custom Summary] Extension(x=1)'
    )
    self.assertEqual(
        view.render(
            TestView.Extension(1), use_content=True).content,
        '[Custom Content] Extension(x=1)'
    )
    self.assertEqual(
        view.render(
            TestView.Extension(1),
            use_summary=True,
            default_behavior=True
        ).content,
        '[Default Summary] abc: Extension(x=1)'
    )
    self.assertEqual(
        base.view(
            TestView.Extension(1),
            view_id='test',
            use_summary=True,
            title='def',
            default_behavior=True,
            content_only=True,
        ).content,
        '[Default Summary] def: Extension(x=1)'
    )

  def test_view_selection_with_multi_inheritance(self):
    self.assertEqual(
        base.view(MyObject(1, 'foo'), view_id='test').content,
        '[Custom Document] MyObject(x=1)'
    )
    self.assertEqual(
        base.view(MyObject(1, 'foo'), view_id='test2').content,
        '<span>1</span>'
    )

  def test_view_with_content_input(self):
    doc = Document('abc')
    self.assertIs(base.view(doc), doc)


if __name__ == '__main__':
  unittest.main()
